r"""Auxiliary functions for reading and writing MetaImage files."""

import io
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union
import zlib

import numpy as np
import torch
from torch import Tensor

from deepali.core.grid import Grid
from deepali.core.pathlib import PathUri, path_suffix
from deepali.core.storage import StorageObject


META_IMAGE_SUFFIXES = (".mha", ".mhd")


def has_meta_image_suffix(path: PathUri) -> bool:
    r"""Check whether filename ends in a MetaImage extension."""
    return path_suffix(path).lower() in META_IMAGE_SUFFIXES


def read_meta_image(arg: Union[PathUri, bytes, io.BufferedReader]) -> Tuple[Tensor, Grid]:
    r"""Read image data and grid attributes from MetaImage file."""
    if isinstance(arg, (Path, str)):
        suffix = path_suffix(arg).lower()
        if suffix == ".mha":
            if isinstance(arg, Path):
                with Path(arg).open("rb") as f:
                    data, meta = read_meta_image_from_fileobj(f)
            else:
                with StorageObject.from_path(arg) as obj:
                    with io.BytesIO(obj.read_bytes()) as f:
                        data, meta = read_meta_image_from_fileobj(f)
        elif suffix == ".mhd":
            try:
                from .sitk import read_sitk_image
            except ImportError:
                raise RuntimeError("SimpleITK is required for reading .mhd MetaImage files")
            return read_sitk_image(arg)
        else:
            raise ValueError(f"MetaImage filename suffix must be one of {META_IMAGE_SUFFIXES}")
    elif isinstance(arg, bytes):
        with io.BytesIO(arg) as f:
            data, meta = read_meta_image_from_fileobj(f)
    elif isinstance(arg, io.BufferedReader):
        data, meta = read_meta_image_from_fileobj(arg)
    else:
        raise TypeError("read_meta_image() 'arg' must be path, bytes, or io.BufferedReader")
    if meta.get("ElementNumberOfChannels", 1) == 1:
        data = np.expand_dims(data, 0)
    else:
        data = np.squeeze(np.swapaxes(np.expand_dims(data, 0), 0, -1), -1)
    size = tuple(meta.get("DimSize", data.shape[:0:-1]))
    assert size[::-1] == data.shape[1:]
    origin = meta.get("Position", meta.get("Origin", meta.get("Offset")))
    matrix = meta.get("Rotation", meta.get("Orientation", meta.get("TransformMatrix")))
    spacing = meta.get("ElementSpacing")
    grid = Grid(
        size=size,
        origin=origin,
        spacing=spacing,
        direction=matrix,
    )
    if data.dtype == np.uint16:
        data = data.astype(np.int32)
    elif data.dtype == np.uint32:
        data = data.astype(np.int64)
    return torch.from_numpy(data), grid


def write_meta_image(data: Tensor, grid: Grid, path: PathUri, compress: bool = True) -> None:
    r"""Write image data and grid attributes to MetaImage file."""
    suffix = path_suffix(path).lower()
    if suffix == ".mha":
        if data.ndim > grid.ndim + 1:
            raise ValueError("write_image() data.ndim must be equal to grid.ndim or grid.ndim + 1")
        meta = {
            "CompressedData": compress,
            "ElementNumberOfChannels": data.shape[0],
            "ElementSpacing": grid.spacing().cpu().numpy(),
            "Offset": grid.origin().cpu().numpy(),
            "TransformMatrix": grid.direction().cpu().numpy(),
        }
        data = data.detach().cpu()
        if data.ndim == grid.ndim + 1:
            if data.shape[0] > 1:
                data = data.unsqueeze(-1).transpose_(0, -1)
            data = data.squeeze(0)
        blob = meta_image_bytes(data.numpy(), meta)
        with StorageObject.from_path(path) as obj:
            obj.write_bytes(blob)
    elif suffix == ".mhd":
        try:
            from .sitk import write_sitk_image
        except ImportError:
            raise RuntimeError("SimpleITK is required for writing .mhd MetaImage files")
        return write_sitk_image(data, grid, path, compress=compress)
    else:
        raise ValueError(f"MetaImage filename suffix must be one of {META_IMAGE_SUFFIXES}")


# ------------------------------------------------------------------------------------------------------------------------------
# The MIT License (MIT)
#
# Copyright (c) 2021 Ali Uneri
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
# IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ------------------------------------------------------------------------------------------------------------------------------

MetaData = Dict[str, Any]

# https://itk.org/Wiki/ITK/MetaIO/Documentation#Reference:_Tags_of_MetaImage
META_IMAGE_TAGS = (
    "Comment",  # MET_STRING
    "ObjectType",  # MET_STRING (Image)
    "ObjectSubType",  # MET_STRING
    "TransformType",  # MET_STRING (Rigid)
    "NDims",  # MET_INT
    "Name",  # MET_STRING
    "ID",  # MET_INT
    "ParentID",  # MET_INT
    "CompressedData",  # MET_STRING (boolean)
    "CompressedDataSize",  # MET_INT
    "BinaryData",  # MET_STRING (boolean)
    "BinaryDataByteOrderMSB",  # MET_STRING (boolean)
    "ElementByteOrderMSB",  # MET_STRING (boolean)
    "Color",  # MET_FLOAT_ARRAY[4]
    "Position",  # MET_FLOAT_ARRAY[NDims]
    "Offset",  # == Position
    "Origin",  # == Position
    "Orientation",  # MET_FLOAT_MATRIX[NDims][NDims]
    "Rotation",  # == Orientation
    "TransformMatrix",  # == Orientation
    "CenterOfRotation",  # MET_FLOAT_ARRAY[NDims]
    "AnatomicalOrientation",  # MET_STRING (RAS)
    "ElementSpacing",  # MET_FLOAT_ARRAY[NDims]
    "DimSize",  # MET_INT_ARRAY[NDims]
    "HeaderSize",  # MET_INT
    "HeaderSizePerSlice",  # MET_INT (non-standard tag for handling per slice header)
    "Modality",  # MET_STRING (MET_MOD_CT)
    "SequenceID",  # MET_INT_ARRAY[4]
    "ElementMin",  # MET_FLOAT
    "ElementMax",  # MET_FLOAT
    "ElementNumberOfChannels",  # MET_INT
    "ElementSize",  # MET_FLOAT_ARRAY[NDims]
    "ElementType",  # MET_STRING (MET_UINT)
    "ElementDataFile",  # MET_STRING
)

META_IMAGE_TYPES = {
    "MET_CHAR": np.int8,
    "MET_UCHAR": np.uint8,
    "MET_SHORT": np.int16,
    "MET_USHORT": np.uint16,
    "MET_INT": np.int32,
    "MET_UINT": np.uint32,
    "MET_LONG": np.int64,
    "MET_ULONG": np.uint64,
    "MET_FLOAT": np.float32,
    "MET_DOUBLE": np.float64,
}


# Adapted from https://github.com/auneri/MetaImageIO/blob/079914fdb55e0720fdb7db53baa5ab20af2fe556/metaimageio/read.py
def read_meta_image_from_fileobj(f: io.BufferedReader) -> Tuple[np.ndarray, MetaData]:
    # read metadata
    meta_in = {}
    meta_size = 0
    islocal = False
    f.seek(0)
    for line in f:
        line = line.decode()
        meta_size += len(line)
        # skip empty and commented lines
        if not line or line.startswith("#"):
            continue
        key = line.split("=", 1)[0].strip()
        value = line.split("=", 1)[-1].strip()
        # handle case variations
        try:
            key = META_IMAGE_TAGS[[x.upper() for x in META_IMAGE_TAGS].index(key.upper())]
        except ValueError:
            pass
        meta_in[key] = value
        # handle supported ElementDataFile formats
        if key == "ElementDataFile":
            if value.upper() != "LOCAL":
                raise NotImplementedError("ElementDataFile must be 'LOCAL'")
            islocal = True
            break
    if not islocal:
        raise ValueError("Missing ElementDataFile header key")

    # typecast metadata to native types
    meta = dict.fromkeys(META_IMAGE_TAGS, None)
    for key, value in meta_in.items():
        if key in (
            "Comment",
            "ObjectType",
            "ObjectSubType",
            "TransformType",
            "Name",
            "AnatomicalOrientation",
            "Modality",
            "ElementDataFile",
        ):
            meta[key] = value
        elif key in (
            "NDims",
            "ID",
            "ParentID",
            "CompressedDataSize",
            "HeaderSize",
            "HeaderSizePerSlice",
            "ElementNumberOfChannels",
        ):
            meta[key] = np.uintp(value)
        elif key in (
            "CompressedData",
            "BinaryData",
            "BinaryDataByteOrderMSB",
            "ElementByteOrderMSB",
        ):
            meta[key] = value.upper() == "TRUE"
        elif key in (
            "Color",
            "Position",
            "Offset",
            "Origin",
            "CenterOfRotation",
            "ElementSpacing",
            "ElementSize",
        ):
            meta[key] = np.array(value.split(), dtype=float)
        elif key in ("Orientation", "Rotation", "TransformMatrix"):
            meta[key] = np.array(value.split(), dtype=float).reshape(3, 3).transpose()
        elif key in ("DimSize", "SequenceID"):
            meta[key] = np.array(value.split(), dtype=int)
        elif key in ("ElementMin", "ElementMax"):
            meta[key] = float(value)
        elif key == "ElementType":
            try:
                meta[key] = [x[1] for x in META_IMAGE_TYPES.items() if x[0] == value.upper()][0]
            except IndexError as exception:
                raise ValueError(f'ElementType "{value}" is not supported') from exception
        else:
            meta[key] = value

    # read image from file
    shape = np.asarray(meta["DimSize"]).copy()[::-1]
    if (meta.get("ElementNumberOfChannels") or 1) > 1:
        shape = np.r_[shape, meta["ElementNumberOfChannels"]]
    else:
        meta["ElementNumberOfChannels"] = 1
    element_size = np.dtype(meta["ElementType"]).itemsize
    increment = np.prod(shape[1:], dtype=int) * element_size

    f.seek(meta_size, 0)
    f.seek((meta.get("HeaderSize") or 0), 1)
    if meta.get("CompressedData"):
        if meta["CompressedDataSize"] is None:
            raise ValueError("CompressedDataSize needs to be specified when using CompressedData")
        if meta["HeaderSizePerSlice"] is not None:
            raise ValueError("HeaderSizePerSlice is not supported with compressed images")
        buffer = zlib.decompress(f.read(meta["CompressedDataSize"]))
    else:
        data = io.BytesIO()
        read, seek = 0, 0
        for _ in range(shape[0]):
            if meta["HeaderSizePerSlice"] is not None:
                data.write(f.read(read))
                read = 0
                seek += meta["HeaderSizePerSlice"]
            f.seek(int(seek), 1)
            seek = 0
            read += increment
            if read > np.iinfo(int).max - increment:
                data.write(f.read(read))
                read = 0
        data.write(f.read(read))
        buffer = data.getbuffer()

    image = np.frombuffer(buffer, dtype=meta["ElementType"]).reshape(shape)
    if meta.get("BinaryDataByteOrderMSB") or meta.get("ElementByteOrderMSB"):
        image.byteswap(inplace=True)
    image = image.copy()

    # remove unused metadata
    meta["ElementDataFile"] = None
    meta = {x: y for x, y in meta.items() if y is not None}

    return image, meta


# Adapted from https://github.com/auneri/MetaImageIO/blob/57cc445afe0560fa50e245d6875b7efff7a1d880/metaimageio/writer.py
def meta_image_bytes(data: np.ndarray, meta: Optional[MetaData] = None) -> bytes:
    r"""Serialize image to bytes object in MetaImage .mha format.

    Args:
        data: NumPy array of (uncompressed) image data.
        meta: Metadata dictionary.

    Returns:
        blob: Uncompressed or compressed MHA file data.

    """

    data = np.asanyarray(data)

    # handle ElementNumberOfChannels
    if meta and meta["ElementNumberOfChannels"] is not None and meta["ElementNumberOfChannels"] > 1:
        size = np.array(data.shape[:-1][::-1])
        ndim = np.ndim(data) - 1
    else:
        size = np.array(data.shape[::-1])
        ndim = np.ndim(data)

    # initialize metadata
    meta_out: MetaData = dict.fromkeys(META_IMAGE_TAGS, None)
    meta_out["ObjectType"] = "Image"
    meta_out["NDims"] = ndim
    meta_out["BinaryData"] = True
    meta_out["BinaryDataByteOrderMSB"] = False
    meta_out["ElementSpacing"] = np.ones(ndim)
    meta_out["DimSize"] = size
    meta_out["ElementType"] = data.dtype

    # input metadata (case incensitive)
    if meta:
        for key, value in meta.items():
            try:
                key = META_IMAGE_TAGS[[x.upper() for x in META_IMAGE_TAGS].index(key.upper())]
            except ValueError:
                pass
            meta_out[key] = value
    meta = meta_out

    # define ElementDataFile
    meta["ElementDataFile"] = "LOCAL"

    # prepare image for saving
    if meta.get("BinaryDataByteOrderMSB") or meta.get("ElementByteOrderMSB"):
        data = data.byteswap()
    blob = data.astype(meta["ElementType"]).tobytes()
    if meta.get("CompressedData", True):
        blob = zlib.compress(blob, level=2)
        meta["CompressedDataSize"] = len(blob)

    # typecast metadata to string
    meta_out: MetaData = {}
    for key, value in meta.items():
        if value is None:
            continue
        elif key in (
            "Comment",
            "ObjectType",
            "ObjectSubType",
            "TransformType",
            "Name",
            "AnatomicalOrientation",
            "Modality",
        ):
            meta_out[key] = value
        elif key in (
            "NDims",
            "ID",
            "ParentID",
            "CompressedData",
            "CompressedDataSize",
            "BinaryData",
            "BinaryDataByteOrderMSB",
            "ElementByteOrderMSB",
            "HeaderSize",
            "HeaderSizePerSlice",
            "ElementMin",
            "ElementMax",
            "ElementNumberOfChannels",
        ):
            meta_out[key] = str(value)
        elif key in (
            "Color",
            "Position",
            "Offset",
            "Origin",
            "CenterOfRotation",
            "ElementSpacing",
            "DimSize",
            "SequenceID",
            "ElementSize",
        ):
            meta_out[key] = " ".join(str(x) for x in np.ravel(value))
        elif key in ("Orientation", "Rotation", "TransformMatrix"):
            meta_out[key] = " ".join(str(x) for x in np.ravel(np.transpose(value)))
        elif key == "ElementType":
            try:
                meta_out[key] = [
                    x[0] for x in META_IMAGE_TYPES.items() if np.issubdtype(value, x[1])
                ][0]
            except IndexError as exception:
                raise ValueError(
                    f"meta_image_bytes() ElementType '{value}' is not supported"
                ) from exception
        elif key == "ElementDataFile":
            if not isinstance(value, str) or value.upper() != "LOCAL":
                raise ValueError("meta_image_bytes() ElementDataFile must be 'LOCAL'")
            meta_out[key] = value
        else:
            meta_out[key] = value
    meta = meta_out

    # serialize meta and image data
    stream = io.BytesIO()
    stream.writelines(f"{key} = {value}\n".encode("ascii") for key, value in meta.items())
    stream.write(blob)
    return stream.getvalue()



r"""Utilities for setting up logging in a main scripts."""

from __future__ import annotations

from argparse import Namespace
from enum import Enum
import logging
from logging import Logger
from typing import Optional, Union


class LogLevel(str, Enum):
    r"""Enumeration of logging levels.

    In particular, this enumeration can be used for type annotation of log level argument when using Typer.

    """

    NOTSET = "NOTSET"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

    @classmethod
    def from_arg(cls, value: Union[int, str, LogLevel, None]) -> LogLevel:
        if value is None:
            return LogLevel.NOTSET
        if isinstance(value, LogLevel):
            return value
        if isinstance(value, int):
            return cls.from_int(value)
        if isinstance(value, str):
            return cls.from_str(value)
        raise TypeError(f"{cls.__name__}.from_arg() 'value' must be int, str, LogLevel, or None")

    @classmethod
    def from_int(cls, value: int) -> LogLevel:
        if not isinstance(value, int):
            raise TypeError(f"{cls.__name__}.from_int() 'value' must be int")
        log_level = LogLevel.NOTSET
        for level in cls:
            if int(level) > value:
                break
            log_level = level
        return log_level

    @classmethod
    def from_str(cls, value: str) -> LogLevel:
        if not isinstance(value, str):
            raise TypeError(f"{cls.__name__}.from_str() 'value' must be str")
        return cls(value.upper())

    @classmethod
    def from_logger(cls, logger: Logger) -> LogLevel:
        return cls.from_int(logger.level)

    def __str__(self) -> str:
        return self.value

    def __int__(self) -> int:
        r"""Cast enumeration to int logging level."""
        return int(getattr(logging, self.value))

    def __eq__(self, other: Union[LogLevel, int]) -> bool:
        return int(self) == int(other)

    def __lt__(self, other: Union[LogLevel, int]) -> bool:
        return int(self) < int(other)

    def __le__(self, other: Union[LogLevel, int]) -> bool:
        return int(self) <= int(other)

    def __gt__(self, other: Union[LogLevel, int]) -> bool:
        return int(self) > int(other)

    def __ge__(self, other: Union[LogLevel, int]) -> bool:
        return int(self) <= int(other)


LOG_FORMAT = "%(asctime)-15s [%(levelname)s] %(message)s"
LOG_LEVELS = tuple(log_level.value for log_level in LogLevel)


def configure_logging(
    logger: Logger,
    args: Optional[Namespace] = None,
    log_level: Optional[Union[int, str, LogLevel]] = None,
    format: Optional[str] = None,
):
    r"""Initialize logging."""
    logging.basicConfig(format=format or LOG_FORMAT)
    if args is not None:
        log_level = getattr(args, "log_level", log_level)
    log_level = LogLevel.from_arg(log_level)
    if log_level is LogLevel.NOTSET:
        log_level = LogLevel.INFO
    logger.setLevel(log_level.value)
    logging.getLogger("botocore").setLevel(logging.ERROR)
    logging.getLogger("s3transfer").setLevel(logging.ERROR)
    logging.getLogger("urllib3").setLevel(logging.ERROR)



r"""Predefined convolutional kernels."""

import math
from typing import Optional, Sequence, Union

import torch
from torch import Tensor
import torch.nn.functional as F

from .linalg import tensordot
from .tensor import as_tensor, cat_scalars
from .typing import Array, Device, DType, Scalar


def bspline1d(stride: int, order: int = 4) -> Tensor:
    r"""B-spline kernel of given order for specified control point spacing.

    Implementation adopted from AirLab:
    https://github.com/airlab-unibas/airlab/blob/80c9d487c012892c395d63c6d937a67303c321d1/airlab/utils/kernelFunction.py#L218

    This function computes the kernel recursively by convolving with a box filter (cf. Cox-de Boor's recursion formula).
    The resulting kernel differs from the analytic B-spline function. This may be due to the box filter having extend
    to the borders of the pixels, where it should drop to zero at pixel centers rather.

    The exact B-spline kernel of order 4 is computed by ``cubic_bspline1d()``.

    Args:
        stride: Spacing between control points with respect to original (upsampled) image grid.
        order: Order of B-spline kernel, where the degree of the spline polynomials is order minus 1.

    Returns:
        B-spline convolution kernel.

    """
    kernel = kernel_ones = torch.ones(1, 1, stride, dtype=torch.float)
    for _ in range(1, order + 1):
        kernel = F.conv1d(kernel, kernel_ones, padding=stride - 1) / stride
    return kernel.reshape(-1)


def cubic_bspline_value(x: float, derivative: int = 0) -> float:
    r"""Evaluate 1-dimensional cubic B-spline."""
    t = abs(x)
    # outside local support region
    if t >= 2:
        return 0
    # 0-th order derivative
    if derivative == 0:
        if t < 1:
            return 2 / 3 + (0.5 * t - 1) * t**2
        return -((t - 2) ** 3) / 6
    # 1st order derivative
    if derivative == 1:
        if t < 1:
            return (1.5 * t - 2.0) * x
        if x < 0:
            return 0.5 * (t - 2) ** 2
        return -0.5 * (t - 2) ** 2
    # 2nd oder derivative
    if derivative == 2:
        if t < 1:
            return 3 * t - 2
        return -t + 2


def cubic_bspline(
    stride: Union[int, Sequence[int]],
    *args: int,
    derivative: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
):
    r"""Get n-dimensional cubic B-spline kernel.

    Args:
        stride: Spacing between control points with respect to original (upsampled) image grid.
        derivative: Order of cubic B-spline derivative.

    Returns:
        Cubic B-spline convolution kernel.

    """
    stride_ = cat_scalars(
        stride,
        *args,
        derivative=derivative,
        dtype=torch.int32,
        device=torch.device("cpu"),
    ).tolist()
    D = len(stride_)
    if D == 1:
        return cubic_bspline1d(stride_, derivative=derivative, dtype=dtype, device=device)
    if D == 2:
        return cubic_bspline2d(stride_, derivative=derivative, dtype=dtype, device=device)
    if D == 3:
        return cubic_bspline3d(stride_, derivative=derivative, dtype=dtype, device=device)
    raise NotImplementedError(f"cubic_bspline() {D}-dimensional kernel")


def cubic_bspline1d(
    stride: Union[int, Sequence[int]],
    derivative: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Cubic B-spline kernel for specified control point spacing.

    Args:
        stride: Spacing between control points with respect to original (upsampled) image grid.
        derivative: Order of cubic B-spline derivative.

    Returns:
        Cubic B-spline convolution kernel.

    """
    if dtype is None:
        dtype = torch.float
    if not isinstance(stride, int):
        (stride,) = stride
    kernel = torch.ones(4 * stride - 1, dtype=torch.float)
    radius = kernel.shape[0] // 2
    for i in range(kernel.shape[0]):
        kernel[i] = cubic_bspline_value((i - radius) / stride, derivative=derivative)
    if device is None:
        device = kernel.device
    return kernel.to(device)


def cubic_bspline2d(
    stride: Union[int, Sequence[int]],
    *args: int,
    derivative: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Cubic B-spline kernel for specified control point spacing.

    Args:
        stride: Spacing between control points with respect to original (upsampled) image grid.
        derivative: Order of cubic B-spline derivative.

    Returns:
        Cubic B-spline convolution kernel.

    """
    if dtype is None:
        dtype = torch.float
    stride_ = cat_scalars(stride, *args, num=2, dtype=torch.int32, device=torch.device("cpu"))
    kernel = torch.ones((4 * stride_ - 1).tolist(), dtype=dtype)
    radius = [n // 2 for n in kernel.shape]
    for j in range(kernel.shape[1]):
        w_j = cubic_bspline_value((j - radius[1]) / stride[1], derivative=derivative)
        for i in range(kernel.shape[0]):
            w_i = cubic_bspline_value((i - radius[0]) / stride[0], derivative=derivative)
            kernel[j, i] = w_i * w_j
    if device is None:
        device = kernel.device
    return kernel.to(device)


def cubic_bspline3d(
    stride: Union[int, Sequence[int]],
    *args: int,
    derivative: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Cubic B-spline kernel for specified control point spacing.

    Args:
        stride: Spacing between control points with respect to original (upsampled) image grid.
        derivative: Order of cubic B-spline derivative.

    Returns:
        Cubic B-spline convolution kernel.

    """
    if dtype is None:
        dtype = torch.float
    stride_ = cat_scalars(stride, *args, num=3, dtype=torch.int32, device=torch.device("cpu"))
    kernel = torch.ones((4 * stride_ - 1).tolist(), dtype=torch.float)
    radius = [n // 2 for n in kernel.shape]
    for k in range(kernel.shape[2]):
        w_k = cubic_bspline_value((k - radius[2]) / stride[2], derivative=derivative)
        for j in range(kernel.shape[1]):
            w_j = cubic_bspline_value((j - radius[1]) / stride[1], derivative=derivative)
            for i in range(kernel.shape[0]):
                w_i = cubic_bspline_value((i - radius[0]) / stride[0], derivative=derivative)
                kernel[k, j, i] = w_i * w_j * w_k
    if device is None:
        device = kernel.device
    return kernel.to(device)


def gaussian_kernel_radius(sigma: Union[Scalar, Array], factor: Scalar = 3) -> Tensor:
    r"""Radius of truncated Gaussian kernel.

    Args:
        sigma: Standard deviation in grid units.
        factor: Number of standard deviations at which to truncate.

    Returns:
        Radius of truncated Gaussian kernel in grid units.

    """
    sigma = as_tensor(sigma, dtype=torch.float32, device="cpu")
    is_scalar = sigma.ndim == 0
    if is_scalar:
        sigma = sigma.unsqueeze(0)
    if sigma.ndim != 1:
        raise ValueError("gaussian() 'sigma' must be scalar or sequence")
    if sigma.shape[0] == 0:
        raise ValueError("gaussian() 'sigma' must be scalar or non-empty sequence")
    if sigma.lt(0).any():
        raise ValueError("Gaussian standard deviation must be non-negative")
    factor = as_tensor(factor, dtype=sigma.dtype, device=sigma.device)
    radius = sigma.mul(factor).floor().type(torch.int64)
    if is_scalar:
        radius = radius
    return radius


def gaussian(
    sigma: Union[Scalar, Array],
    *args: Scalar,
    normalize: bool = True,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
):
    r"""Get n-dimensional Gaussian kernel."""
    sigma_ = cat_scalars(sigma, *args, dtype=dtype, device=torch.device("cpu"))
    if not torch.is_floating_point(sigma_):
        if dtype is not None:
            raise TypeError("Gaussian kernel dtype must be floating point type")
        sigma_ = sigma_.type(torch.float)
    if sigma_.ndim == 0:
        sigma_ = sigma_.unsqueeze(0)
    if sigma_.ndim != 1:
        raise ValueError("gaussian() 'sigma' must be scalar or sequence")
    if sigma_.shape[0] == 0:
        raise ValueError("gaussian() 'sigma' must be scalar or non-empty sequence")
    kernel = gaussian1d(sigma_[0], normalize=False, dtype=torch.float64)
    for std in sigma_[1:]:
        other = gaussian1d(std, normalize=False, dtype=torch.float64)
        kernel = tensordot(kernel, other, dims=0)
    if normalize:
        kernel /= kernel.sum()
    return kernel.to(dtype=sigma_.dtype, device=device)


def gaussian1d(
    sigma: Scalar,
    radius: Optional[Union[int, Tensor]] = None,
    scale: Optional[Scalar] = None,
    normalize: bool = True,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Get 1-dimensional Gaussian kernel."""
    sigma = as_tensor(sigma, device="cpu")
    if sigma.ndim != 0:
        raise ValueError("gaussian1d() 'sigma' must be scalar")
    if sigma.lt(0):
        raise ValueError("Gaussian standard deviation must be non-negative")
    if dtype is not None and not dtype.is_floating_point:
        raise TypeError("Gaussian kernel dtype must be floating point type")
    if radius is None:
        radius = gaussian_kernel_radius(sigma)
    radius = int(radius)
    if radius > 0:
        size = 2 * radius + 1
        x = torch.linspace(-radius, radius, steps=size, dtype=dtype, device=device)
        sigma = sigma.to(dtype=dtype, device=device)
        kernel = torch.exp(-0.5 * ((x / sigma) ** 2))
        if scale is None:
            scale = 1 / sigma.mul(math.sqrt(2 * math.pi))
        else:
            scale = as_tensor(scale, dtype=dtype, device=device)
        kernel *= scale
        if normalize:
            kernel /= kernel.sum()
    else:
        if scale is None:
            scale = 1
        kernel = as_tensor([scale], dtype=dtype, device=device)
    return kernel


def gaussian1d_I(
    sigma: Scalar,
    normalize: bool = True,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Get 1st order derivative of 1-dimensional Gaussian kernel."""
    if torch.is_tensor(sigma):
        sigma_ = as_tensor(sigma)  # satisfy static type checker
        if sigma_.ndim != 0:
            raise ValueError("gaussian1d() 'sigma' must be scalar")
        sigma = sigma_.item()
    if sigma < 0:
        raise ValueError("Gaussian standard deviation must be non-negative")
    if dtype is not None and not dtype.is_floating_point:
        raise TypeError("Gaussian kernel dtype must be floating point type")
    radius = int(gaussian_kernel_radius(sigma).item())
    if radius > 0:
        size = 2 * radius + 1
        x = torch.linspace(-radius, radius, steps=size, dtype=dtype, device=device)
        norm = torch.tensor(1 / (sigma * math.sqrt(2 * math.pi)), dtype=dtype, device=device)
        var = sigma**2
        # Note that conv1d() computes correlation, i.e., the kernel is mirrored
        kernel = norm * torch.exp(-0.5 * x**2 / var) * (x / var)
        if normalize:
            kernel /= (norm * torch.exp(-0.5 * x**2 / var)).sum()
    else:
        kernel = torch.tensor([1], dtype=dtype, device=device)
    return kernel


def gaussian2d(sigma: Union[Scalar, Array], *args: Scalar, **kwargs) -> Tensor:
    r"""Get 2-dimensional Gaussian kernel."""
    sigma = cat_scalars(sigma, *args, num=2, device=torch.device("cpu"))
    return gaussian(sigma, **kwargs)


def gaussian3d(sigma: Union[Scalar, Array], *args: Scalar, **kwargs) -> Tensor:
    r"""Get 3-dimensional Gaussian kernel."""
    sigma = cat_scalars(sigma, *args, num=3, device=torch.device("cpu"))
    return gaussian(sigma, **kwargs)



r"""Auxiliary functions for torch.nn.Module instances."""

from collections import OrderedDict, defaultdict
from typing import Any, Dict, List, Mapping, Optional

from torch.nn import Module


def has_children(module: Module) -> bool:
    r"""Check if module has other modules as children."""
    try:
        next(iter(module.children()))
    except StopIteration:
        return False
    return True


def module_ids_by_class_name(
    model: Module,
    duplicates: bool = False,
    containers: bool = False,
    memo: Optional[dict] = None,
) -> Dict[str, List[int]]:
    r"""Obtain ids of module objects in model, indexed by module type name.

    Args:
        model: Neural network model.
        duplicates: Whether to return lists of object ids instead of sets.
        containers: Whether to include object ids of modules with children.
        memo: Used internally to recursively collect object ids.

    Returns:
        Dictionary with module class names as keys and object ids of each type stored list.

    """
    if memo is None:
        memo = defaultdict(list)
    modules: dict = model._modules
    for module in modules.values():
        assert isinstance(module, Module)
        if containers or not has_children(module):
            ids = memo[module.__class__.__name__]
            ids.append(id(module))
        module_ids_by_class_name(module, duplicates=duplicates, containers=containers, memo=memo)
    if not duplicates:
        memo = {k: list(set(v)) for k, v in memo.items()}
    return memo


def module_counts_by_class_name(
    model: Module, duplicates: bool = False, containers: bool = False
) -> Dict[str, int]:
    r"""Count how many module objects are in a model of each module type.

    Args:
        model: Neural network model.
        duplicates: Whether to count duplicate objects.
        containers: Whether to include modules with children.

    Returns:
        Dictionary with module class names as keys and counts of object ids as values.

    """
    ids = module_ids_by_class_name(model, duplicates=duplicates, containers=containers)
    return {k: len(v) for k, v in ids.items()}


def remove_layers_in_state_dict(state: Mapping[str, Any], prefix: str) -> Dict[str, Any]:
    r"""Remove layers in loaded state dict."""
    metadata = getattr(state, "_metadata", None)
    result = OrderedDict(state)  # does not copy _metadata attribute
    if metadata is not None:
        metadata = remove_layers_in_state_dict(metadata, prefix)
        setattr(result, "_metadata", metadata)
    for key in state.keys():
        if key == prefix:
            result.pop(key)
        elif key.startswith(prefix + "."):
            result.pop(key)
    return result


def rename_layers_in_state_dict(
    state: Mapping[str, Any],
    rename: Optional[Mapping[str, str]] = None,
    parent: Optional[str] = None,
) -> Dict[str, Any]:
    r"""Rename layers in loaded state dict."""
    metadata = getattr(state, "_metadata", None)
    state = OrderedDict(state)  # does not copy _metadata attribute
    if metadata is not None:
        metadata = rename_layers_in_state_dict(metadata, rename=rename, parent=parent)
        setattr(state, "_metadata", metadata)
    if rename:
        for prefix, new_name in rename.items():
            for key in list(state.keys()):
                if key == prefix:
                    value = state.pop(key)
                    if new_name:
                        state[new_name] = value
                elif key.startswith(prefix + "."):
                    value = state.pop(key)
                    new_key = new_name + key[len(prefix) + (0 if new_name else 1) :]
                    if new_key:
                        state[new_key] = value
    if parent:
        state = {f"{parent}.{key}": value for key, value in state.items()}
    return state



r"""Functions for B-spline interpolation."""

from typing import Callable, Optional, Sequence, Tuple, Union, overload

import torch
from torch import Size, Tensor
from torch.nn import functional as F

from .enum import PaddingMode, SpatialDim, SpatialDimArg
from .grid import Grid
from .image import conv, conv1d
from .kernels import cubic_bspline1d
from .tensor import move_dim
from .typing import DType, ScalarOrTuple


@overload
def cubic_bspline_control_point_grid_size(size: int, stride: int) -> int:
    ...


@overload
def cubic_bspline_control_point_grid_size(size: Sequence[int], stride: int) -> Tuple[int, ...]:
    ...


@overload
def cubic_bspline_control_point_grid_size(size: int, stride: Sequence[int]) -> Tuple[int, ...]:
    ...


@overload
def cubic_bspline_control_point_grid_size(
    size: Sequence[int], stride: Sequence[int]
) -> Tuple[int, ...]:
    ...


def cubic_bspline_control_point_grid_size(
    size: ScalarOrTuple[int], stride: ScalarOrTuple[int]
) -> ScalarOrTuple[int]:
    r"""Calculate required number of cubic B-spline coefficients for given output size."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError(
            "cubic_bspline_control_point_grid_size() 'size' must be scalar or sequence"
        )
    if m.le(0).any():
        raise ValueError("cubic_bspline_control_point_grid_size() 'size' must be positive")
    if s.le(0).any():
        raise ValueError("cubic_bspline_control_point_grid_size() 'stride' must be positive")
    ndim = m.shape[0]
    if ndim == 1 and s.shape[0] > 1:
        ndim = s.shape[0]
    for arg, name in zip([m, s], ["size", "stride"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"cubic_bspline_control_point_grid_size() {name!r} must be scalar or sequence of length {ndim}"
            )
    m = m.expand(ndim)
    s = s.expand(ndim)
    n = m.div(s, rounding_mode="floor").add_(3)
    n = n.where(m % s == 0, n.add(1))
    if isinstance(size, int) and isinstance(stride, int):
        return n[0].item()
    return Size(n.tolist())


def cubic_bspline_control_point_grid(grid: Grid, stride: ScalarOrTuple[int]) -> Grid:
    r"""Get control point grid for given image grid and control point stride."""
    size = cubic_bspline_control_point_grid_size(grid.size(), stride)
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=grid.device))
    s = s.expand(grid.ndim)
    return Grid(
        size=size,
        origin=grid.index_to_world(-s),
        spacing=grid.spacing(),
        direction=grid.direction(),
        device=grid.device,
        align_corners=True,
    )


@overload
def bspline_interpolation_weights(
    degree: int,
    stride: int,
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Tensor:
    ...


@overload
def bspline_interpolation_weights(
    degree: int,
    stride: Sequence[int],
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Tuple[Tensor, ...]:
    ...


def bspline_interpolation_weights(
    degree: int,
    stride: ScalarOrTuple[int],
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Union[Tensor, Tuple[Tensor, ...]]:
    r"""Compute B-spline interpolation weights."""
    # Adapted from MIRTK ComputeBSplineIndicesAndWeights() at
    # https://github.com/BioMedIA/MIRTK/blob/877917b1b3ec9e602f7395dbbb1270e4334fc311/Modules/Numerics/include/mirtk/BSpline.h#L670-L789
    if degree == 3:
        return cubic_bspline_interpolation_weights(stride, dtype=dtype, device=device)
    kernels = {}
    return_single_tensor = False
    if isinstance(stride, int):
        stride = [stride]
        return_single_tensor = True
    for s in stride:
        if s in kernels:
            continue
        kernel = torch.empty((s, degree + 1), dtype=dtype, device=device)
        offset = torch.arange(0, 1, 1 / s, dtype=kernel.dtype, device=kernel.device)
        if degree % 2 == 0:
            offset = offset.sub_(offset.round())
        if degree == 2:
            kernel[:, 1] = 0.75 - offset.square()
            kernel[:, 2] = offset.sub(kernel[:, 1]).add_(1).mul_(0.5)
            kernel[:, 0] = -kernel[:, [1, 2]].sum(1).sub(1)
        # degree == 3: See cubic_bspline_interpolation_weights()
        elif degree == 4:
            # MIRTK code variable names: w=offset, w2=a
            a = offset.square()
            t = a.mul(1 / 6)
            t0 = t.sub(11 / 24).mul(offset)
            t1 = t.sub(0.25).mul_(-a).add_(19 / 96)
            kernel[:, 0] = torch.tensor(0.5, dtype=dtype, device=device).sub(offset).square()
            kernel[:, 0] = kernel[:, 0].mul(kernel[:, 0].mul(1 / 24))
            kernel[:, 1] = t1.add(t0)
            kernel[:, 3] = t1.sub(t0)
            kernel[:, 4] = offset.mul(0.5).add_(kernel[:, 0]).add_(t0)
            kernel[:, 2] = -kernel[:, [0, 1, 3, 4]].sum(1).sub(1)
        elif degree == 5:
            # MIRTK code variable names: w=offset, w2=a, w4=b
            a = offset.square()
            kernel[:, 5] = offset.mul(a.square()).mul_(1 / 120)
            a = a.sub_(offset)
            b = a.square()
            offset = offset.sub_(0.5)
            t = a.sub(3).mul_(a)
            kernel[:, 0] = a.add(b).add_(1 / 5).mul_(1 / 24).sub_(kernel[:, 5])
            t0 = a.sub(5).mul_(a).add_(46 / 5).mul_(1 / 24)
            t1 = t.add(4).mul_(offset).mul_(-1 / 12)
            kernel[:, 2] = t0.add(t1)
            kernel[:, 3] = t0.sub(t1)
            t0 = t.sub(9 / 5).mul_(1.0 / 16.0)
            t1 = b.sub(a).sub_(5).mul_(offset).mul_(1.0 / 24.0)
            kernel[:, 1] = t0.add(t1)
            kernel[:, 4] = t0.sub(t1)
        else:
            raise NotImplementedError(f"B-spline interpolation for degree={degree}")
        kernels[s] = kernel
    kernels = tuple(kernels[s] for s in stride)
    if return_single_tensor:
        assert len(kernels) == 1
        return kernels[0]
    return kernels


@overload
def cubic_bspline_interpolation_weights(
    stride: int,
    derivative: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Tensor:
    ...


@overload
def cubic_bspline_interpolation_weights(
    stride: int,
    derivative: Sequence[int],
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Tuple[Tensor, ...]:
    ...


@overload
def cubic_bspline_interpolation_weights(
    stride: Sequence[int],
    derivative: ScalarOrTuple[int] = 0,
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Tuple[Tensor, ...]:
    ...


def cubic_bspline_interpolation_weights(
    stride: ScalarOrTuple[int],
    derivative: ScalarOrTuple[int] = 0,
    dtype: Optional[DType] = None,
    device: Optional[torch.device] = None,
) -> Union[Tensor, Tuple[Tensor, ...]]:
    r"""Compute cubic B-spline interpolation weights."""
    kernels = {}
    return_single_tensor = isinstance(stride, int) and isinstance(derivative, int)
    if isinstance(stride, int):
        stride = [stride] * (len(derivative) if isinstance(derivative, Sequence) else 1)
    if isinstance(derivative, int):
        derivative = [derivative] * len(stride)
    elif not isinstance(derivative, Sequence):
        raise TypeError(
            "cubic_bspline_interpolation_weights() 'derivative' must be int or Sequence[int]"
        )
    elif len(derivative) != len(stride):
        raise ValueError(
            "cubic_bspline_interpolation_weights() length of 'derivative' sequence does not match 'stride'"
        )
    for s, d in zip(stride, derivative):
        if (s, d) in kernels:
            continue
        kernel = torch.empty((s, 4), dtype=dtype, device=device)
        offset = torch.arange(0, 1, 1 / s, dtype=kernel.dtype, device=kernel.device)
        if d == 0:
            kernel[:, 3] = offset.pow(3).mul_(1 / 6)
            kernel[:, 0] = offset.mul(offset.sub(1)).mul_(0.5).add_(1 / 6).sub_(kernel[:, 3])
            kernel[:, 2] = offset.add(kernel[:, 0]).sub_(kernel[:, 3].mul(2))
            kernel[:, 1] = -kernel[:, [0, 2, 3]].sum(1).sub(1)
        elif d == 1:
            kernel[:, 3] = offset.pow(2).mul_(0.5)
            kernel[:, 0] = offset.sub(kernel[:, 3]).sub_(0.5)
            kernel[:, 2] = kernel[:, 0].sub(kernel[:, 3].mul(2)).add_(1)
            kernel[:, 1] = -kernel[:, [0, 2, 3]].sum(1)
        elif d == 2:
            kernel[:, 3] = offset
            kernel[:, 0] = -offset.sub(1)
            kernel[:, 2] = -offset.mul(3).sub_(1)
            kernel[:, 1] = offset.mul(3).sub_(2)
        elif d == 3:
            kernel[:, 3] = 1
            kernel[:, 0] = -1
            kernel[:, 2] = -3
            kernel[:, 1] = 3
        else:
            kernel.fill_(0)
        kernels[(s, d)] = kernel
    kernels = tuple(kernels[(s, d)] for s, d in zip(stride, derivative))
    if return_single_tensor:
        assert len(kernels) == 1
        return kernels[0]
    return kernels


def evaluate_cubic_bspline(
    data: Tensor,
    stride: Optional[ScalarOrTuple[int]] = None,
    size: Optional[Size] = None,
    shape: Optional[Size] = None,
    kernel: Optional[Union[Tensor, Sequence[Tensor]]] = None,
    derivative: Optional[ScalarOrTuple[int]] = None,
    transpose: bool = False,
) -> Tensor:
    r"""Evaluate cubic B-spline function.

    Args:
        data: Cubic B-spline interpolation coefficients as tensor of shape ``(N, C, ..., X)``.
        stride: Number of output grid points between control points plus one. This is the stride of the
            transposed convolution used to upsample the control point displacements to the output size.
            If a sequence of values is given, these must be the strides for the different spatial
            dimensions in the order ``(sx, ...)``. When a ``kernel`` is specified and ``transpose=False``,
            this argument is ignored. Otherwise it is required.
        size: Spatial size of output tensor in the order ``(nx, ...)``.
        shape: Spatial size of output tensor in the order ``(..., nx)``.
        kernel: Precomputed cubic B-spline interpolation kernel. When multiple 1D kernels are given,
            these must be in the order ``(kx, ...)``.
        derivative: Order of partial derivatives along each spatial dimension. When a scalar is given,
            the same order is applied to all spatial dimensions. When a ``kernel`` is specified and
            ``transpose=False``, this argument is ignored. For ``transpose=True``, the evaluation of
            partial derivatives is not implemented, and hence this argument must be None or 0.
        transpose: Whether to use separable transposed convolution as implemented in AIRLab.
            When ``False``, a more efficient implementation using multi-channel convolution followed
            by a reshuffling of the output is performed. This more efficient and also more accurate
            implementation is adapted from the C++ code of MIRTK (``mirtk::BSplineInterpolateImageFunction``).

    Returns:
        Cubic B-spline function values as tensor of shape ``(N, C, ..., X')``, where ``X' = sx * X``
        when neither output ``size`` nor ``shape`` is specified. Otherwise, the output tensor is cropped
        to the requested spatial output size.

    """
    if not isinstance(data, Tensor):
        raise TypeError("evaluate_cubic_bspline() 'data' must be torch.Tensor")
    if not torch.is_floating_point(data):
        raise TypeError("evaluate_cubic_bspline() 'data' must have floating point dtype")
    if data.ndim < 3:
        raise ValueError("evaluate_cubic_bspline() 'data' must have shape (N, C, ..., X)")
    if size is not None:
        if shape is not None:
            raise ValueError("evaluate_cubic_bspline() 'size' and 'shape' are mutually exclusive")
        shape = Size(reversed(size))
    D = data.ndim - 2
    N = data.shape[0]
    C = data.shape[1]
    # Implementation inspired by AIRLab
    if transpose:
        if stride is None:
            raise ValueError("evaluate_cubic_bspline() 'stride' is required when transpose=True")
        if isinstance(stride, int):
            stride = [stride] * D
        if any(s < 1 for s in stride):
            raise ValueError("evaluate_cubic_bspline() 'stride' must be positive")
        if kernel is None:
            if (isinstance(derivative, int) and derivative != 0) or (
                isinstance(derivative, Sequence) and any(order != 0 for order in derivative)
            ):
                raise NotImplementedError(
                    "evaluate_cubic_bspline() 'derivative' order cannot be non-zero when kernel=None and transpose=True"
                )
            kernels = {}
            for s in stride:
                if s not in kernels:
                    kernels[s] = cubic_bspline1d(s)
            kernel = [kernels[s] for s in stride]
        stride = tuple(reversed(stride))
        if isinstance(kernel, Sequence):
            kernel = tuple(reversed(kernel))
        output = conv(
            data,
            kernel=kernel,
            stride=stride,
            padding=PaddingMode.ZEROS,
            transpose=True,
        )
        if shape is not None:
            output = output[
                (slice(0, N), slice(0, C)) + tuple(slice(s, s + n) for s, n in zip(stride, shape))
            ]
    # Implementation adapted from MIRTK
    else:
        if kernel is None:
            if stride is None:
                stride = 1
            if isinstance(stride, int):
                stride = [stride] * D
            if any(s < 1 for s in stride):
                raise ValueError("evaluate_cubic_bspline() 'stride' must be positive")
            if derivative is None:
                derivative = 0
            kernel = cubic_bspline_interpolation_weights(
                stride=stride,
                derivative=derivative,
                dtype=data.dtype,
                device=data.device,
            )
        elif isinstance(kernel, Tensor):
            kernel = [kernel] * D
        elif not isinstance(kernel, Sequence):
            raise TypeError(
                "evaluate_cubic_bspline() 'kernel' must be Tensor or Sequence of tensors"
            )
        output = data
        dims = tuple(SpatialDim(dim).tensor_dim(data.ndim) for dim in range(D))
        conv_fn: Callable[..., Tensor] = [F.conv1d, F.conv2d, F.conv3d][D - 1]
        for dim, w in zip(dims, kernel):
            if w.ndim == 1:
                w = w.unsqueeze(0)
            if w.ndim != 2:
                raise ValueError(
                    "evaluate_cubic_bspline() 'kernel' must be 1- or 2-dimensional tensors"
                )
            weight = w.reshape((w.shape[0], 1, w.shape[1]) + (1,) * (D - 1))
            weight = weight.tile((C,) + (1,) * (weight.ndim - 1))
            output = move_dim(output, dim, 2)
            output = conv_fn(output, weight, groups=C)
            output = output.reshape((N, C, w.shape[0]) + (output.shape[2:]))
            output = output.transpose(2, 3).flatten(2, 3)
            output = move_dim(output, 2, dim)
        if shape is not None:
            output = output[(slice(0, N), slice(0, C)) + tuple(slice(0, n) for n in shape)]
    return output


def subdivide_cubic_bspline(
    data: Tensor, dims: Optional[Union[SpatialDimArg, Sequence[SpatialDimArg]]] = None
) -> Tensor:
    r"""Compute cubic B-spline coefficients for subdivided control point grid.

    Args:
        data: Input control point coefficients as tensor of shape ``(N, C, ..., X)``.
        dims: Spatial dimensions along which to subdivide.

    Returns:
        Coefficients of subdivided cubic B-spline function.

    """
    if not isinstance(data, Tensor):
        raise TypeError("subdivide_cubic_bspline() 'data' must be torch.Tensor")
    if not torch.is_floating_point(data):
        raise TypeError("subdivide_cubic_bspline() 'data' must have floating point dtype")
    if data.ndim < 4:
        raise ValueError("subdivide_cubic_bspline() 'data' must have shape (N, C, ..., X)")
    if dims is None:
        dims = tuple(range(data.ndim - 2))
    elif isinstance(dims, (int, str)):
        dims = [dims]
    elif not isinstance(dims, Sequence):
        raise TypeError("subdivide_cubic_bspline() 'dims' must be int, str, or Sequence thereof")
    dims = sorted(SpatialDim.from_arg(dim).tensor_dim(data.ndim) for dim in dims)
    output = data
    kernel_1 = torch.tensor([0.125, 0.75, 0.125], dtype=data.dtype, device=data.device)
    kernel_2 = torch.tensor([0.5, 0.5], dtype=data.dtype, device=data.device)
    for dim in dims:
        # Allocate tensor for subdivided control point coefficients
        shape = output.shape[:dim] + (2 * output.shape[dim] - 1,) + output.shape[dim + 1 :]
        temp = torch.empty(shape, dtype=data.dtype, device=data.device)
        # Evaluate coefficients at original control point positions
        indices = [slice(0, n) for n in shape]
        indices[dim] = slice(0, shape[dim], 2)
        temp[indices] = conv1d(output, kernel_1, dim=dim, padding=1)
        # Evaluate coefficients at subdivided intermediate positions
        indices = [slice(0, n) for n in shape]
        indices[dim] = slice(1, shape[dim], 2)
        temp[indices] = conv1d(output, kernel_2, dim=dim, padding=0)
        output = temp
    return output



r"""Modules for sampling of image data, e.g., after spatial transformation."""

from typing import Dict, Mapping, Optional, Tuple, Union, cast, overload

from torch import Tensor
from torch.nn import Module

from deepali.core import functional as U
from deepali.core.enum import PaddingMode, Sampling
from deepali.core.grid import Axes, Grid, grid_points_transform
from deepali.core.linalg import homogeneous_matmul, homogeneous_transform
from deepali.core.typing import Scalar


class SampleImage(Module):
    r"""Sample images at grid points."""

    def __init__(
        self,
        target: Grid,
        source: Optional[Grid] = None,
        axes: Optional[Union[Axes, str]] = None,
        sampling: Optional[Union[Sampling, str]] = None,
        padding: Optional[Union[PaddingMode, str, Scalar]] = None,
        align_centers: bool = False,
    ):
        r"""Initialize base class.

        Args:
            target: Grid on which to sample transformed images.
            source: Grid on which input source images are sampled.
            axes: Axes with respect to which grid coordinates are defined.
                If ``None``, use cube axes corresponding to ``target.align_corners()`` setting.
            sampling: Image interpolation mode.
            padding: Image extrapolation mode.
            align_centers: Whether to implicitly align the ``target`` and ``source`` centers.
                If ``True``, only the affine component of the target to source transformation
                is applied. If ``False``, also the translation of grid center points is considered.

        """
        super().__init__()
        self._target = target
        self._source = source or target
        if axes is None:
            axes = Axes.from_grid(target)
        self._axes = Axes(axes)
        self._sampling = Sampling.from_arg(sampling)
        if padding is None or isinstance(padding, (PaddingMode, str)):
            self._padding = PaddingMode.from_arg(padding)
        else:
            self._padding = float(padding)
        self._align_centers = bool(align_centers)
        # Precompute target points to source cube transformation **AFTER** attributes are set!
        self.register_buffer("matrix", self._matrix())

    def axes(self) -> Axes:
        r"""Axes with respect to which target grid points and transformations thereof are defined."""
        return self._axes

    def target_grid(self) -> Grid:
        r"""Target sampling grid."""
        return self._target

    def source_grid(self) -> Grid:
        r"""Source sampling grid."""
        return self._source

    def sampling(self) -> Sampling:
        r"""Image sampling mode."""
        return self._sampling

    def padding(self) -> Union[PaddingMode, Scalar]:
        r"""Image padding mode or value, respectively."""
        return self._padding

    def padding_mode(self) -> PaddingMode:
        r"""Image padding mode."""
        return self._padding if isinstance(self._padding, PaddingMode) else PaddingMode.CONSTANT

    def padding_value(self) -> float:
        r"""Image padding value if mode is "constant"."""
        return 0.0 if isinstance(self._padding, PaddingMode) else float(self._padding)

    def align_centers(self) -> bool:
        r"""Whether grid center points are implicitly aligned."""
        return self._align_centers

    def align_corners(self) -> bool:
        r"""Whether to sample images using ``align_corners=False`` or ``align_corners=True``."""
        return self._target.align_corners()

    def _matrix(self) -> Tensor:
        r"""Homogeneous coordinate transformation from target grid points to source grid cube."""
        align_corners = self.align_corners()
        to_axes = Axes.from_align_corners(align_corners)
        matrix = grid_points_transform(self._target, self._axes, self._source, to_axes)
        if self._align_centers:
            offset = self._target.world_to_cube(self._source.center(), align_corners=align_corners)
            matrix = homogeneous_matmul(matrix, offset)
        return matrix.unsqueeze(0)

    def _transform_target_to_source(self, grid: Tensor) -> Tensor:
        r"""Transform target grid points to source cube."""
        matrix = cast(Tensor, self.matrix)
        return homogeneous_transform(matrix, grid)

    def _sample_source_image(
        self,
        grid: Tensor,
        input: Optional[Union[Tensor, Mapping[str, Tensor]]] = None,
        data: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor, Tensor], Dict[str, Tensor]]:
        r"""Sample images at specified source grid points."""
        source = {}
        output = {}
        shape = None
        align_corners = self.align_corners()
        if not isinstance(grid, Tensor):
            raise TypeError(f"{type(self).__name__}() 'grid' must be Tensor")
        if isinstance(input, dict):
            if data is not None or mask is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'input' dict and 'data'/'mask' are mutually exclusive"
                )
            for data in input.values():
                if not isinstance(data, Tensor):
                    raise TypeError(f"{type(self).__name__}() 'input' dict values must be Tensor")
            source: Dict[str, Tensor] = {
                name: data for name, data in input.items() if name != "mask"
            }
            shape = next(iter(source.values())).shape if source else None
            mask = input.get("mask")
        elif isinstance(input, Tensor):
            if data is not None:
                raise ValueError(
                    f"{type(self).__name__}() 'input' and 'data' are mutually exclusive"
                )
            source = {"data": input}
            shape = input.shape
        elif input is None:
            if data is None and mask is None:
                raise ValueError(
                    f"{type(self).__name__} 'input', 'data', and/or 'mask' is required"
                )
        else:
            raise TypeError(
                f"{type(self).__name__}() 'input' must be Tensor or Mapping[str, Tensor]"
            )
        for name, data in source.items():
            is_unbatched = data.ndim == grid.shape[-1] + 1
            if is_unbatched:
                data = data.unsqueeze(0)
            data = U.grid_sample(
                data,
                grid,
                mode=self._sampling,
                padding=self._padding,
                align_corners=align_corners,
            )
            if is_unbatched:
                data = data.squeeze(0)
            output[name] = data
        if mask is not None:
            if not isinstance(mask, Tensor):
                raise TypeError(f"{type(self).__name__}() 'mask' must be Tensor")
            if shape is not None:
                if mask.ndim != len(shape):
                    raise ValueError(
                        f"{type(self).__name__}() 'mask' must have same ndim as 'input' data"
                    )
                if mask.shape[0] != shape[0]:
                    raise ValueError(
                        f"{type(self).__name__}() 'mask' must have same batch size as 'input' data"
                    )
                if mask.shape[2:] != shape[2:]:
                    raise ValueError(
                        f"{type(self).__name__}() 'mask' must have same spatial shape as 'input' data"
                    )
            temp = U.grid_sample_mask(mask, grid, align_corners=align_corners)
            output["mask"] = temp > 0.9
        if isinstance(input, dict):
            return output
        if data is None:
            return output["mask"]
        if mask is None:
            return output["data"]
        return output["data"], output["mask"]

    @overload
    def forward(self, grid: Tensor, input: Tensor, data=None, mask=None) -> Tensor:
        r"""Sample batch of images at spatially transformed target grid points."""
        ...

    @overload
    def forward(self, grid: Tensor, input: Dict[str, Tensor]) -> Dict[str, Tensor]:
        r"""Sample batch of optionally masked images at spatially transformed target grid points."""
        ...

    def forward(
        self,
        grid: Tensor,
        input: Optional[Union[Tensor, Dict[str, Tensor]]] = None,
        data: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor, Tensor], Dict[str, Tensor]]:
        r"""Sample images at target grid points after mapping these to the source grid cube."""
        if grid.ndim == grid.shape[-1] + 1:
            grid = grid.unsqueeze(0)
        grid = self._transform_target_to_source(grid)
        return self._sample_source_image(grid, input=input, data=data, mask=mask)

    def extra_repr(self) -> str:
        return (
            f"target={repr(self._target)}"
            + f", source={repr(self._source)}"
            + f", axes={repr(self._axes.value)}"
            + f", sampling={repr(self._sampling.value)}"
            + f", padding={repr(self._padding.value if isinstance(self._padding, PaddingMode) else self._padding)}"
            + f", align_centers={repr(self._align_centers)}"
        )


class TransformImage(SampleImage):
    r"""Sample images at transformed target grid points.

    This module can be used for both linear and non-rigid transformations, where the shape of the
    input ``transform`` tensor determines the type of transformation. See also ``transform_grid()``.

    """

    def __init__(
        self,
        target: Grid,
        source: Optional[Grid] = None,
        axes: Optional[Union[Axes, str]] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str, Scalar] = PaddingMode.BORDER,
        align_centers: bool = False,
    ):
        r"""Initialize module.

        Args:
            target: Grid on which to sample transformed images.
            source: Grid on which input source images are sampled.
            axes: Axes with respect to which transformations are defined.
                Use ``Axes.from_grid(target)`` if ``None``.
            sampling: Image interpolation mode.
            padding: Image extrapolation mode.
            align_centers: Whether to implicitly align the ``target`` and ``source`` centers.
                If ``True``, only the affine component of the target to source transformation
                is applied. If ``False``, also the translation of grid center points is considered.

        """
        super().__init__(
            target,
            source,
            axes=axes,
            sampling=sampling,
            padding=padding,
            align_centers=align_centers,
        )
        self.register_buffer("grid", self._grid(), persistent=False)

    def _grid(self) -> Tensor:
        r"""Target grid points before spatial transformation."""
        return self._target.points(self._axes).unsqueeze(0)

    def forward(
        self,
        transform: Optional[Tensor],
        input: Optional[Union[Tensor, Dict[str, Tensor]]] = None,
        data: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor, Tensor], Dict[str, Tensor]]:
        r"""Sample images at transformed target grid points after mapping these to the source grid cube."""
        grid = cast(Tensor, self.grid)
        if isinstance(transform, Tensor):
            if transform.ndim == grid.shape[-1] + 1:
                transform = transform.unsqueeze(0)
            grid = U.transform_grid(transform, grid, align_corners=self.align_corners())
        elif transform is not None:
            raise TypeError("TransformImage() 'transform' must be Tensor")
        grid = self._transform_target_to_source(grid)
        return self._sample_source_image(grid, input=input, data=data, mask=mask)


class AlignImage(SampleImage):
    r"""Sample images at linearly transformed target grid points.

    Instead of applying two separate linear transformations to the target grid points, this module first composes
    the two linear transformations and then applies the composite transformation to the target grid points.

    """

    def __init__(
        self,
        target: Grid,
        source: Optional[Grid] = None,
        axes: Optional[Union[Axes, str]] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str, Scalar] = PaddingMode.BORDER,
        align_centers: bool = False,
    ):
        r"""Initialize module.

        Args:
            target: Grid on which to sample transformed images.
            source: Grid on which input source images are sampled.
            axes: Axes with respect to which transformations are defined.
                Use ``Axes.from_grid(target)`` if ``None``.
            sampling: Image interpolation mode.
            padding: Image extrapolation mode.
            align_centers: Whether to implicitly align the ``target`` and ``source`` centers.
                If ``True``, only the affine component of the target to source transformation
                is applied. If ``False``, also the translation of grid center points is considered.

        """
        super().__init__(
            target,
            source,
            axes=axes,
            sampling=sampling,
            padding=padding,
            align_centers=align_centers,
        )
        self.register_buffer("grid", self._grid(), persistent=False)

    def _grid(self) -> Tensor:
        r"""Target grid points before spatial transformation."""
        return self._target.points(self._axes).unsqueeze(0)

    def forward(
        self,
        transform: Optional[Tensor],
        input: Optional[Union[Tensor, Dict[str, Tensor]]] = None,
        data: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor, Tensor], Dict[str, Tensor]]:
        r"""Sample batch of optionally masked images at linearly transformed target grid points."""
        composite_transform = cast(Tensor, self.matrix)
        if transform is not None:
            if not isinstance(transform, Tensor):
                raise TypeError("AlignImage() 'transform' must be Tensor")
            if transform.ndim != 3:
                raise ValueError("AlignImage() 'transform' must be 3-dimensional tensor")
            composite_transform = homogeneous_matmul(composite_transform, transform)
        grid = cast(Tensor, self.grid)
        grid = homogeneous_transform(composite_transform, grid)
        return self._sample_source_image(grid, input=input, data=data, mask=mask)



r"""Access most core library functions with a single import.

.. code-block:: python

    import deepali.core.functional as U

"""

# Basic tensor functions
from .math import abspow
from .math import atanh
from .math import max_difference
from .math import round_decimals
from .math import threshold

from .tensor import as_tensor
from .tensor import as_float_tensor
from .tensor import as_one_hot_tensor
from .tensor import atleast_1d
from .tensor import batched_index_select
from .tensor import move_dim
from .tensor import unravel_coords
from .tensor import unravel_index

# Linear algebra/geometry
from .affine import affine_rotation_matrix
from .affine import apply_transform as apply_affine_transform
from .affine import euler_rotation_matrix
from .affine import euler_rotation_angles
from .affine import euler_rotation_order
from .affine import identity_transform
from .affine import rotation_matrix
from .affine import scaling_transform
from .affine import shear_matrix
from .affine import translation
from .affine import transform_points as affine_transform_points
from .affine import transform_vectors as affine_transform_vectors

from .linalg import angle_axis_to_rotation_matrix
from .linalg import angle_axis_to_quaternion
from .linalg import as_homogeneous_matrix
from .linalg import as_homogeneous_tensor
from .linalg import hmm
from .linalg import homogeneous_matmul
from .linalg import homogeneous_matrix
from .linalg import homogeneous_transform
from .linalg import normalize_quaternion
from .linalg import quaternion_to_angle_axis
from .linalg import quaternion_to_rotation_matrix
from .linalg import quaternion_log_to_exp
from .linalg import quaternion_exp_to_log
from .linalg import rotation_matrix_to_angle_axis
from .linalg import rotation_matrix_to_quaternion
from .linalg import tensordot
from .linalg import vectordot
from .linalg import vector_rotation

# Data operations
from .bspline import bspline_interpolation_weights
from .bspline import cubic_bspline_control_point_grid
from .bspline import cubic_bspline_control_point_grid_size
from .bspline import evaluate_cubic_bspline
from .bspline import subdivide_cubic_bspline

from .image import avg_pool, max_pool, min_pool
from .image import conv, conv1d
from .image import dot_batch, dot_channels
from .image import downsample, upsample
from .image import gaussian_pyramid
from .image import crop, pad
from .image import center_crop, center_pad
from .image import fill_border
from .image import finite_differences
from .image import flatten_channels
from .image import grid_resample
from .image import grid_reshape
from .image import grid_resize
from .image import grid_sample
from .image import grid_sample_mask
from .image import image_slice
from .image import normalize_image
from .image import rand_sample
from .image import rescale
from .image import sample_image
from .image import spatial_derivatives

from .image import circle_image
from .image import cshape_image
from .image import empty_image
from .image import grid_image
from .image import ones_image
from .image import zeros_image

from .flow import affine_flow
from .flow import compose_flows
from .flow import compose_svfs
from .flow import curl
from .flow import denormalize_flow
from .flow import divergence
from .flow import divergence_free_flow
from .flow import expv
from .flow import flow_derivatives
from .flow import jacobian_det
from .flow import jacobian_dict
from .flow import jacobian_matrix
from .flow import lie_bracket
from .flow import logv
from .flow import normalize_flow
from .flow import sample_flow
from .flow import warp_grid
from .flow import warp_image
from .flow import warp_points
from .flow import zeros_flow

from .pointset import bounding_box
from .pointset import distance_matrix
from .pointset import closest_point_distances
from .pointset import closest_point_indices
from .pointset import denormalize_grid
from .pointset import normalize_grid
from .pointset import polyline_directions
from .pointset import polyline_tangents
from .pointset import transform_grid
from .pointset import transform_points

from .random import multinomial


__all__ = (
    # Basic tensor functions
    "abspow",
    "as_tensor",
    "as_float_tensor",
    "as_one_hot_tensor",
    "atanh",
    "atleast_1d",
    "batched_index_select",
    "max_difference",
    "move_dim",
    "round_decimals",
    "threshold",
    "unravel_coords",
    "unravel_index",
    # Random sampling
    "multinomial",
    # Linear algebra/geometry
    "affine_flow",
    "affine_rotation_matrix",
    "affine_transform_points",
    "affine_transform_vectors",
    "angle_axis_to_rotation_matrix",
    "angle_axis_to_quaternion",
    "apply_affine_transform",
    "as_homogeneous_matrix",
    "as_homogeneous_tensor",
    "euler_rotation_matrix",
    "euler_rotation_angles",
    "euler_rotation_order",
    "hmm",
    "homogeneous_matmul",
    "homogeneous_matrix",
    "homogeneous_transform",
    "identity_transform",
    "normalize_quaternion",
    "quaternion_to_angle_axis",
    "quaternion_to_rotation_matrix",
    "quaternion_log_to_exp",
    "quaternion_exp_to_log",
    "rotation_matrix",
    "rotation_matrix_to_angle_axis",
    "rotation_matrix_to_quaternion",
    "scaling_transform",
    "shear_matrix",
    "tensordot",
    "translation",
    "vectordot",
    "vector_rotation",
    # Data operations
    "avg_pool",
    "bounding_box",
    "bspline_interpolation_weights",
    "center_crop",
    "center_pad",
    "circle_image",
    "closest_point_distances",
    "closest_point_indices",
    "compose_flows",
    "compose_svfs",
    "conv",
    "conv1d",
    "crop",
    "cshape_image",
    "cubic_bspline_control_point_grid",
    "cubic_bspline_control_point_grid_size",
    "curl",
    "denormalize_flow",
    "denormalize_grid",
    "distance_matrix",
    "divergence",
    "divergence_free_flow",
    "dot_batch",
    "dot_channels",
    "downsample",
    "empty_image",
    "evaluate_cubic_bspline",
    "expv",
    "flatten_channels",
    "finite_differences",
    "flow_derivatives",
    "gaussian_pyramid",
    "grid_image",
    "image_slice",
    "pad",
    "fill_border",
    "grid_resample",
    "grid_reshape",
    "grid_resize",
    "grid_sample",
    "grid_sample_mask",
    "jacobian_det",
    "jacobian_dict",
    "jacobian_matrix",
    "lie_bracket",
    "logv",
    "max_pool",
    "min_pool",
    "normalize_flow",
    "normalize_grid",
    "normalize_image",
    "ones_image",
    "polyline_directions",
    "polyline_tangents",
    "rand_sample",
    "rescale",
    "sample_flow",
    "sample_image",
    "spatial_derivatives",
    "subdivide_cubic_bspline",
    "transform_grid",
    "transform_points",
    "upsample",
    "warp_grid",
    "warp_image",
    "warp_points",
    "zeros_flow",
    "zeros_image",
)



from torchinfo import summary

from deepali.networks.resnet import ResNet


def test_resnet() -> None:
    r"""Test construction of ResNet model."""

    # ResNet-34 for ImageNet classification
    model = ResNet.from_depth(model_depth=34, spatial_dims=2, in_channels=3, num_classes=1000)
    summary(model, input_size=(1, 3, 256, 256), depth=10, device="cpu")



r"""Losses for image, point set, and/or surface registration.

Classes representing loss terms defined by this package are derived from ``torch.nn.Module``
and follow a stateful object oriented design. The underlying functions implementing these loss
terms are defined in the :mod:`.functional` module following a stateless functional API.

The following import statement can be used to access the functional API:

.. code-block:: python

    import deepali.losses.functional as L

"""

import sys
from typing import Any

from torch.nn import Module

from .base import BSplineLoss
from .base import DisplacementLoss
from .base import NormalizedPairwiseImageLoss
from .base import PairwiseImageLoss
from .base import ParamsLoss
from .base import PointSetDistance
from .base import RegistrationLoss
from .base import RegistrationLosses
from .base import RegistrationResult

from .bspline import BSplineBending, BSplineBendingEnergy

from .flow import Bending, BendingEnergy, BE
from .flow import Curvature
from .flow import Diffusion
from .flow import Divergence
from .flow import Elasticity
from .flow import TotalVariation, TV

from .image import PatchwiseImageLoss, PatchLoss
from .image import Dice, DSC
from .image import NCC
from .image import LCC, LNCC
from .image import SLCC, WLCC
from .image import MI, NMI
from .image import MAE, L1ImageLoss
from .image import HuberImageLoss
from .image import SmoothL1ImageLoss
from .image import MSE, L2ImageLoss
from .image import SSD

from .params import L1Norm, L1_Norm
from .params import L2Norm, L2_Norm
from .params import Sparsity

from .pointset import ClosestPointDistance, CPD
from .pointset import LandmarkPointDistance, LPD


__all__ = (
    # Base types
    "BSplineLoss",
    "DisplacementLoss",
    "NormalizedPairwiseImageLoss",
    "PairwiseImageLoss",
    "PatchwiseImageLoss",
    "PatchLoss",
    "ParamsLoss",
    "PointSetDistance",
    "RegistrationLoss",
    "RegistrationLosses",
    "RegistrationResult",
    "is_pairwise_image_loss",
    "is_displacement_loss",
    "is_pointset_distance",
    # Loss types
    "BE",
    "Bending",
    "BendingEnergy",
    "BSplineBending",
    "BSplineBendingEnergy",
    "ClosestPointDistance",
    "CPD",
    "Curvature",
    "Dice",
    "Diffusion",
    "Divergence",
    "DSC",
    "Elasticity",
    "HuberImageLoss",
    "L1ImageLoss",
    "L2ImageLoss",
    "L1Norm",
    "L1_Norm",
    "L2Norm",
    "L2_Norm",
    "LandmarkPointDistance",
    "LPD",
    "LCC",
    "LNCC",
    "MAE",
    "MI",
    "MSE",
    "NCC",
    "NMI",
    "SLCC",
    "SmoothL1ImageLoss",
    "Sparsity",
    "SSD",
    "TotalVariation",
    "TV",
    "WLCC",
    # Factory function
    "create_loss",
    "new_loss",
)


def is_pairwise_image_loss(arg: Any) -> bool:
    r"""Check if given argument is name or instance of pairwise image loss."""
    return is_loss_of_type(PairwiseImageLoss, arg)


def is_displacement_loss(arg: Any) -> bool:
    r"""Check if given argument is name or instance of displacement field loss."""
    return is_loss_of_type(DisplacementLoss, arg)


def is_pointset_distance(arg: Any) -> bool:
    r"""Check if given argument is name or instance of point set distance."""
    return is_loss_of_type(PointSetDistance, arg)


def is_loss_of_type(base, arg: Any) -> bool:
    r"""Check if given argument is name or instance of pairwise image loss."""
    cls = None
    if isinstance(arg, str):
        cls = getattr(sys.modules[__name__], arg, None)
    elif type(arg) is type:
        cls = arg
    elif arg is not None:
        cls = type(arg)
    if cls is not None:
        bases = list(cls.__bases__)
        while bases:
            b = bases.pop()
            if b is base:
                return True
            bases.extend(b.__bases__)
    return False


def new_loss(name: str, *args, **kwargs) -> Module:
    r"""Initialize new loss module.

    Args:
        name: Name of loss type.
        args: Loss arguments.
        kwargs: Loss keyword arguments.

    Returns:
        New loss module.

    """
    cls = getattr(sys.modules[__name__], name, None)
    if cls is None:
        raise ValueError(f"new_loss() unknown loss {name}")
    if cls is Module or not issubclass(cls, Module):
        raise TypeError(f"new_loss() '{name}' is not a subclass of torch.nn.Module")
    return cls(*args, **kwargs)


create_loss = new_loss



r"""Low-level tensor utility functions."""

from logging import Logger
from typing import Callable, Optional, Tuple, Union

import torch
from torch import Tensor

from .typing import Array, Device, DType, Scalar


def as_tensor(
    arg: Union[Scalar, Array], dtype: Optional[DType] = None, device: Optional[Device] = None
) -> Tensor:
    r"""Create tensor from array if argument is not of type torch.Tensor.

    Unlike ``torch.as_tensor()``, this function preserves the tensor device if ``device=None``.

    """
    if device is None and isinstance(arg, Tensor):
        device = arg.device
    return torch.as_tensor(arg, dtype=dtype, device=device)  # type: ignore


def as_float_tensor(arr: Array) -> Tensor:
    r"""Create tensor with floating point type from argument if it is not yet."""
    arr_ = as_tensor(arr)
    if not torch.is_floating_point(arr_):
        return arr_.type(torch.float)
    return arr_


def as_one_hot_tensor(
    tensor: Tensor,
    num_classes: int,
    ignore_index: Optional[int] = None,
    dtype: Optional[DType] = None,
) -> Tensor:
    r"""Converts label image to one-hot encoding of multi-class segmentation.

    Adapted from: https://github.com/wolny/pytorch-3dunet

    Args:
        tensor: Input tensor of shape ``(N, 1, ..., X)`` or ``(N, C, ..., X)``.
            When a tensor with ``C == num_classes`` is given, it is converted to the specified
            ``dtype`` but not modified otherwise. Otherwise, the input tensor must contain
            class labels in a single channel.
        num_classes: Number of channels/labels.
        ignore_index: Ignore index to be kept during the expansion. The locations of the index
            value in the GT image is stored in the corresponding locations across all channels so
            that this location can be ignored across all channels later e.g. in Dice computation.
            This argument must be ``None`` if ``tensor`` has ``C == num_channels``.
        dtype: Data type of output tensor. Default is ``torch.float``.

    Returns:
        Output tensor of shape ``(N, C, ..., X)``.

    """
    if dtype is None:
        dtype = torch.float
    if not isinstance(tensor, Tensor):
        raise TypeError("as_one_hot_tensor() 'tensor' must be torch.Tensor")
    if tensor.dim() < 3:
        raise ValueError("as_one_hot_tensor() 'tensor' must have shape (N, C, ..., X)")
    if tensor.shape[1] == num_classes:
        return tensor.to(dtype=dtype)
    elif tensor.shape[1] != 1:
        raise ValueError(
            f"as_one_hot_tensor() 'tensor' must have shape (N, 1|{num_classes}, ..., X)"
        )
    # create result tensor shape (NxCxDxHxW)
    shape = list(tensor.shape)
    shape[1] = num_classes
    # scatter to get the one-hot tensor
    if ignore_index is None:
        return torch.zeros(shape, dtype=dtype).to(tensor.device).scatter_(1, tensor, 1)
    # create ignore_index mask for the result
    mask = tensor.expand(shape) == ignore_index
    # clone the src tensor and zero out ignore_index in the inputs
    inputs = tensor.clone()
    inputs[inputs == ignore_index] = 0
    # scatter to get the one-hot tensor
    result = torch.zeros(shape, dtype=dtype).to(inputs.device).scatter_(1, inputs, 1)
    # bring back the ignore_index in the result
    result[mask] = ignore_index
    return result


def atleast_1d(
    arr: Array, dtype: Optional[DType] = None, device: Optional[Device] = None
) -> Tensor:
    r"""Convert array-like argument to 1- or more-dimensional PyTorch tensor."""
    arr_ = as_tensor(arr, dtype=dtype, device=device)
    return arr_.unsqueeze(0) if arr_.ndim == 0 else arr_


def cat_scalars(
    arg: Union[Scalar, Array],
    *args: Scalar,
    num: int = 0,
    dtype: Optional[DType] = None,
    device: Optional[Device] = None,
) -> Tensor:
    r"""Join arguments into single 1-dimensional tensor.

    This auxiliary function is used by ``Grid``, ``Image``, and ``ImageBatch`` to support
    method arguments for different spatial dimensions as either scalar constant, list
    of scalar ``*args``, or single ``Array`` argument. If a single argument of type ``Array``
    is given, it must be a sequence of scalar values.

    Args:
        arg: Either a single scalar or sequence of scalars. If the argument is a ``Tensor``,
            it is cloned and detached in order to avoid unintended side effects.
        args: Additional scalars. If ``arg`` is a sequence, ``args`` must be empty.
        num: Number of expected scalar values. If a single scalar ``arg`` is given,
            it is repeated ``num`` times to create a 1-dimensional array. If ``num=0``,
            the length of the returned array corresponds to the number of given scalars.
        dtype: Data type of output tensor.
        device: Device on which to store tensor.

    Returns:
        Scalar arguments joined into a 1-dimensional tensor.

    """
    if args:
        if isinstance(arg, (tuple, list)) or isinstance(arg, Tensor):
            raise ValueError("arg and args must either be all scalars, or args empty")
        arg = torch.tensor((arg,) + args, dtype=dtype, device=device)
    else:
        arg = as_tensor(arg, dtype=dtype, device=device)
    if arg.ndim == 0:
        arg = arg.unsqueeze(0)
    if arg.ndim != 1:
        if num > 0:
            raise ValueError(f"Expected one scalar, a sequence of length {num}, or {num} args")
        raise ValueError("Expected one scalar, a sequence of scalars, or multiple scalars")
    if num > 0:
        if len(arg) == 1:
            arg = arg.repeat(num)
        elif len(arg) != num:
            raise ValueError(f"Expected one scalar, a sequence of length {num}, or {num} args")
    return arg


def batched_index_select(input: Tensor, dim: int, index: Tensor) -> Tensor:
    r"""Batched version of torch.index_select().

    See https://discuss.pytorch.org/t/batched-index-select/9115/9.

    """
    for i in range(1, len(input.shape)):
        if i != dim:
            index = index.unsqueeze(i)
    shape = list(input.shape)
    shape[0] = -1
    shape[dim] = -1
    index = index.expand(shape)
    return torch.gather(input, dim, index)


def move_dim(tensor: Tensor, dim: int, pos: int) -> Tensor:
    r"""Move the specified tensor dimension to another position."""
    if dim < 0:
        dim = tensor.ndim + dim
    if pos < 0:
        pos = tensor.ndim + pos
    if pos == dim:
        return tensor
    if dim < pos:
        pos += 1
    tensor = tensor.unsqueeze(pos)
    if pos <= dim:
        dim += 1
    tensor = tensor.transpose(dim, pos).squeeze(dim)
    return tensor


def unravel_coords(indices: Tensor, size: Tuple[int, ...]) -> Tensor:
    r"""Converts flat indices into unraveled grid coordinates.

    Args:
        indices: A tensor of flat indices with shape ``(..., N)``.
        size: Sampling grid size with order ``(X, ...)``.

    Returns:
        Grid coordinates of corresponding grid points.

    """
    size = torch.Size(size)
    numel = size.numel()
    if indices.ge(numel).any():
        raise ValueError(f"unravel_coords() indices must be smaller than {numel}")
    coords = torch.zeros(indices.size() + (len(size),), dtype=indices.dtype, device=indices.device)
    for i, n in enumerate(size):
        coords[..., i] = indices % n
        indices = indices // n
    return coords


def unravel_index(indices: Tensor, shape: Tuple[int, ...]) -> Tensor:
    r"""Converts flat indices into unraveled coordinates in a target shape.

    This is a `torch` implementation of `numpy.unravel_index`, but returning a
    tensor of shape (..., N, D) rather than a D-dimensional tuple. See also
    https://github.com/pytorch/pytorch/issues/35674#issuecomment-739492875.

    Args:
        indices: A tensor of indices with shape (..., N).
        shape: The targeted tensor shape of length D.

    Returns:
        Unraveled coordinates as tensor of shape (..., N, D) with coordinates
        in the same order as the input ``shape`` dimensions.

    """
    shape = torch.Size(shape)
    numel = shape.numel()
    if indices.ge(numel).any():
        raise ValueError(f"unravel_coords() indices must be smaller than {numel}")
    coords = torch.zeros(indices.size() + (len(shape),), dtype=indices.dtype, device=indices.device)
    for i, n in enumerate(reversed(shape)):
        coords[..., i] = indices % n
        indices = indices // n
    return coords.flip(-1)


def log_grad_hook(name: str, logger: Optional[Logger] = None) -> Callable[[Tensor], None]:
    r"""Backward hook to print tensor gradient information for debugging."""

    def printer(grad: Tensor) -> None:
        if grad.nelement() == 1:
            msg = f"{name}.grad: value={grad}"
        else:
            msg = (
                f"{name}.grad: shape={tuple(grad.shape)}"
                f", max={grad.max()}, min={grad.min()}"
                f", mean={grad.mean()}"
            )
        if logger is None:
            print(msg)
        else:
            logger.debug(msg)

    return printer


def register_backward_hook(
    tensor: Tensor, hook: Callable[[Tensor], None], retain_grad: bool = False
) -> Tensor:
    r"""Register backward hook and optionally enable retaining gradient."""
    if tensor.requires_grad:
        if retain_grad:
            tensor.retain_grad()
        tensor.register_hook(hook)
    return tensor



r"""Basic tensor operations."""

from typing import Any, Mapping, Optional, Sequence, Union

from torch import Tensor
from torch.nn import Module

from deepali.core import functional as U
from deepali.core.enum import PaddingMode
from deepali.core.typing import ScalarOrTuple


class GetItem(Module):
    r"""Get item at specified input tensor sequence index or with given dictionary key."""

    def __init__(self, key: Any) -> None:
        r"""Set item index or key.

        Args:
            key: Index of item in sequence of input tensors or key into input map.

        """
        super().__init__()
        self.key = key

    def forward(self, input: Union[Sequence[Tensor], Mapping[Any, Tensor]]) -> Tensor:
        return input[self.key]

    def extra_repr(self) -> str:
        return repr(self.key)


class Narrow(Module):
    r"""Narrowed version of input tensor."""

    def __init__(self, dim: int, start: int, length: int) -> None:
        super().__init__()
        self.dim = dim
        self.start = start
        self.length = length

    def forward(self, x: Tensor) -> Tensor:
        return x.narrow(self.dim, self.start, self.length)

    def extra_repr(self) -> str:
        return f"dim={self.dim}, start={self.start}, length={self.length}"


class Pad(Module):
    r"""Pad tensor."""

    def __init__(
        self,
        margin: Optional[ScalarOrTuple[int]] = None,
        padding: Optional[ScalarOrTuple[int]] = None,
        mode: Union[PaddingMode, str] = PaddingMode.ZEROS,
        value: float = 0,
    ) -> None:
        if margin is None and padding is None:
            raise AssertionError("Pad() either 'margin' or 'padding' is required")
        if margin is not None and padding is not None:
            raise AssertionError("Pad() 'margin' and 'padding' are mutually exclusive")
        super().__init__()
        self.margin = margin
        self.padding = padding
        self.mode = PaddingMode(mode)
        self.value = value

    def forward(self, x: Tensor) -> Tensor:
        return U.pad(x, margin=self.margin, num=self.padding, mode=self.mode, value=self.value)

    def extra_repr(self) -> str:
        if self.margin is None:
            s = f"padding={self.padding}"
        else:
            s = f"margin={self.margin}"
        s += f", mode={self.mode.value!r}"
        if self.mode is PaddingMode.CONSTANT:
            s += f", value={self.value}"
        return s


class Reshape(Module):
    r"""Reshape input tensor.

    This module provides a view of the input tensor without making a copy if possible.
    Otherwise, a copy is made of the input data. See ``torch.reshape()`` for details.

    """

    def __init__(self, shape: Sequence[int]) -> None:
        r"""Set output tensor shape.

        Args:
            shape: Output tensor shape, optionally excluding first batch dimension.

        """
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        shape = self.shape
        if len(shape) == x.ndim - 1:
            shape = (-1,) + shape
        return x.reshape(shape)

    def extra_repr(self) -> str:
        return repr(self.shape)


class View(Module):
    r"""View input tensor with specified shape.

    See https://pytorch.org/docs/stable/tensor_view.html for when it is possible to create
    a view of a tensor without reshaping it. For a more general tensor reshape operation,
    see module ``Reshape``.

    """

    def __init__(self, shape: Sequence[int]) -> None:
        r"""Set output tensor shape.

        Args:
            shape: Output tensor shape, optionally excluding first batch dimension.

        """
        super().__init__()
        self.shape = shape

    def forward(self, x: Tensor) -> Tensor:
        shape = self.shape
        if len(shape) == x.ndim - 1:
            shape = (-1,) + shape
        return x.view(*shape)

    def extra_repr(self) -> str:
        return repr(self.shape)



r"""Spatial coordinate and input data transformation modules.

.. hint::

    The spatial transforms defined by this library can be used to implement co-registration
    approaches based on traditional optimization as well as those based on machine learning
    (amortized optimization).

A spatial transformation maps points from a target domain defined with respect to the unit cube
of a target sampling grid, to points defined with respect to the same domain, i.e., the domain
and codomain of the spatial coordinate map are identical. In order to transform an image defined
with respect to a different sampling grid, this transformation has to be followed by a mapping
from target cube domain to source cube domain. This is done, for example, by the spatial transformer
implemented by :class:`.ImageTransformer`.

The ``forward()`` method of a :class:`.SpatialTransform` can be used to spatially transform any set
of points defined with respect to the grid domain of the spatial transformation, including in particular
a tensor of shape ``(N, M, D)``, i.e., a batch of ``N`` point sets with cardinality ``M``. It can
also be applied to a tensor of grid points of shape ``(N, ..., X, D)`` regardless if the grid points
are located at the undeformed grid positions or those of an already deformed grid. In case of a
non-rigid deformation, the point displacements are by default sampled at the input points. The sampled
flow vectors :math:`u` are then added to the input points :math:`x`, producing the output :math:`y = x + u(x)`.
If the boolean flag ``grid=True`` is passed to the :meth:`.SpatialTransform.forward` function, it is assumed
that the coordinates correspond to the positions of undeformed spatial grid points with domain equal to the
domain of the transformation. In this special case, a simple interpolation to resize the vector field to the
size of the input tensor is used. In case of a linear transformation, :math:`y = Ax + t`.

The coordinate domain is :attr:`.Axes.CUBE_CORNERS` if ``grid.align_corners() == True`` (default),
and :attr:`.Axes.CUBE` otherwise.

"""

import sys
from typing import Any, Optional

from deepali.core.grid import Grid

# Base classes for type comparison and annotation
from .base import LinearTransform  # noqa
from .base import NonRigidTransform  # noqa
from .base import ReadOnlyParameters  # noqa
from .base import SpatialTransform

# Parametric transformation mix-in
from .parametric import ParametricTransform  # noqa

# Composite coordinate transformations
from .composite import CompositeTransform  # noqa
from .composite import MultiLevelTransform  # noqa
from .composite import SequentialTransform  # noqa

# Configurable generic transformation
from .generic import TransformConfig  # noqa
from .generic import GenericSpatialTransform  # noqa
from .generic import affine_first  # noqa
from .generic import has_affine_component  # noqa
from .generic import has_nonrigid_component  # noqa
from .generic import nonrigid_components  # noqa
from .generic import transform_components  # noqa

# Elemental linear transformations
from .linear import EulerRotation
from .linear import HomogeneousTransform
from .linear import QuaternionRotation
from .linear import AnisotropicScaling
from .linear import IsotropicScaling  # noqa
from .linear import Shearing
from .linear import Translation  # noqa

# Composite linear transformations
from .linear import RigidTransform
from .linear import RigidQuaternionTransform
from .linear import SimilarityTransform
from .linear import AffineTransform
from .linear import FullAffineTransform

# Non-rigid deformations
from .nonrigid import DenseVectorFieldTransform  # noqa
from .nonrigid import DisplacementFieldTransform
from .nonrigid import StationaryVelocityFieldTransform

# Free-form deformations
from .bspline import BSplineTransform  # noqa
from .bspline import FreeFormDeformation
from .bspline import StationaryVelocityFreeFormDeformation

# Spatial transformers based on a coordinate transformation
from .image import ImageTransform  # noqa (deprecated)
from .transformer import SpatialTransformer  # noqa
from .transformer import ImageTransformer  # noqa
from .transformer import PointSetTransformer  # noqa


# Aliases
Affine = AffineTransform
AffineWithShearing = FullAffineTransform
Disp = DisplacementFieldTransform
DispField = DisplacementFieldTransform
DDF = DisplacementFieldTransform
DVF = DisplacementFieldTransform
FFD = FreeFormDeformation
FullAffine = FullAffineTransform
MatrixTransform = HomogeneousTransform
Quaternion = QuaternionRotation
Rigid = RigidTransform
RigidQuaternion = RigidQuaternionTransform
Rotation = EulerRotation
Scaling = AnisotropicScaling
ShearTransform = Shearing
Similarity = SimilarityTransform
SVF = StationaryVelocityFieldTransform
SVField = StationaryVelocityFieldTransform
SVFFD = StationaryVelocityFreeFormDeformation


LINEAR_TRANSFORMS = (
    "Affine",
    "AffineTransform",
    "AffineWithShearing",
    "AnisotropicScaling",
    "BSplineTransform",
    "EulerRotation",
    "IsotropicScaling",
    "FullAffine",
    "FullAffineTransform",
    "HomogeneousTransform",
    "MatrixTransform",
    "Quaternion",
    "QuaternionRotation",
    "Rigid",
    "RigidTransform",
    "RigidQuaternion",
    "RigidQuaternionTransform",
    "Rotation",
    "Scaling",
    "Shearing",
    "ShearTransform",
    "Similarity",
    "SimilarityTransform",
    "Translation",
)

NONRIGID_TRANSFORMS = (
    "Disp",
    "DispField",
    "DisplacementFieldTransform",
    "DDF",
    "DVF",
    "FFD",
    "FreeFormDeformation",
    "StationaryVelocityFieldTransform",
    "StationaryVelocityFreeFormDeformation",
    "SVF",
    "SVField",
    "SVFFD",
)

COMPOSITE_TRANSFORMS = (
    "MultiLevelTransform",
    "SequentialTransform",
)

__all__ = (
    (
        "CompositeTransform",
        "DenseVectorFieldTransform",
        "GenericSpatialTransform",
        "ImageTransform",
        "ImageTransformer",
        "LinearTransform",
        "NonRigidTransform",
        "ParametricTransform",
        "PointSetTransformer",
        "ReadOnlyParameters",
        "SpatialTransform",
        "SpatialTransformer",
        "TransformConfig",
        "affine_first",
        "has_affine_component",
        "has_nonrigid_component",
        "is_linear_transform",
        "is_nonrigid_transform",
        "is_spatial_transform",
        "new_spatial_transform",
        "nonrigid_components",
        "transform_components",
    )
    + COMPOSITE_TRANSFORMS
    + LINEAR_TRANSFORMS
    + NONRIGID_TRANSFORMS
)


def is_spatial_transform(arg: Any) -> bool:
    r"""Whether given object or named transformation is a transformation type.

    Args:
        arg: Name of type or object.

    Returns:
        Whether type of ``arg`` object or name of type is a transformation model.

    """
    if isinstance(arg, str):
        return arg in LINEAR_TRANSFORMS or arg in NONRIGID_TRANSFORMS
    return isinstance(arg, SpatialTransform)


def is_linear_transform(arg: Any) -> bool:
    r"""Whether given object is a linear transformation type.

    Args:
        arg: Name of type or object.

    Returns:
        Whether type of ``arg`` object or name of type is a linear transformation.

    """
    if isinstance(arg, str):
        return arg in LINEAR_TRANSFORMS
    if isinstance(arg, SpatialTransform):
        return arg.linear
    return False


def is_nonrigid_transform(arg: Any) -> bool:
    r"""Whether given object is a non-rigid transformation type.

    Args:
        arg: Name of type or object.

    Returns:
        Whether type of ``arg`` object or name of type is a non-rigid transformation.

    """
    if isinstance(arg, str):
        return arg in NONRIGID_TRANSFORMS
    if isinstance(arg, SpatialTransform):
        return arg.nonrigid
    return False


def new_spatial_transform(
    name: str, grid: Grid, groups: Optional[int] = None, **kwargs
) -> SpatialTransform:
    r"""Initialize new transformation model of named type.

    Args:
        name: Name of transformation model.
        grid: Grid of transformation domain.
        groups: Number of transformations.
        kwargs: Optional keyword arguments of transformation model.

    Returns:
        New transformation module with optimizable parameters.

    """
    cls = getattr(sys.modules[__name__], name, None)
    if cls is not None and (name in LINEAR_TRANSFORMS or name in NONRIGID_TRANSFORMS):
        return cls(grid, groups=groups, **kwargs)
    raise ValueError(f"new_spatial_transform() 'name={name}' is not a valid transformation type")



from typing import Sequence

import torch.optim
from torch.optim import Optimizer
from torch.nn import Module


def new_optimizer(name: str, model: Module, **kwargs) -> Optimizer:
    r"""Initialize new optimizer for parameters of given model.

    Args:
        name: Name of optimizer.
        model: Module whose parameters are to be optimized.
        kwargs: Keyword arguments for named optimizer.

    Returns:
        New optimizer instance.

    """
    cls = getattr(torch.optim, name, None)
    if cls is None:
        raise ValueError(f"Unknown optimizer: {name}")
    if not issubclass(cls, Optimizer):
        raise TypeError(f"Requested type '{name}' is not a subclass of torch.optim.Optimizer")
    if "learning_rate" in kwargs:
        if "lr" in kwargs:
            raise ValueError("new_optimizer() 'lr' and 'learning_rate' are mutually exclusive")
        kwargs["lr"] = kwargs.pop("learning_rate")
    return cls(model.parameters(), **kwargs)


def slope_of_least_squares_fit(values: Sequence[float]) -> float:
    r"""Compute slope of least squares fit of line to last n objective function values

    See also:
    - https://www.che.udel.edu/pdf/FittingData.pdf
    - https://en.wikipedia.org/wiki/1_%2B_2_%2B_3_%2B_4_%2B_%E2%8B%AF
    - https://proofwiki.org/wiki/Sum_of_Sequence_of_Squares

    """
    n = len(values)
    if n < 2:
        return float("nan")
    if n == 2:
        return values[1] - values[0]
    # sum_x1 divided by n as a slight modified to reduce no. of operations,
    # i.e., the other terms are divided by n as well by dropping one factor n
    sum_x1 = (n + 1) / 2
    sum_x2 = n * (n + 1) * (2 * n + 1) / 6
    sum_y1 = sum(values)
    sum_xy = sum(((x + 1) * y for x, y in enumerate(values)))
    return (sum_xy - sum_x1 * sum_y1) / (sum_x2 - n * sum_x1 * sum_x1)



r"""Join features from separate network paths."""

from typing import Callable, Sequence, Union

import torch
from torch import Tensor

from .lambd import LambdaLayer


JoinFunc = Callable[[Sequence[Tensor]], Tensor]


def join_func(arg: Union[str, JoinFunc], dim: int = 1) -> JoinFunc:
    r"""Tensor operation which combines features of input tensors, e.g., along skip connection.

    Args:
        arg: Name of operation: 'add': Elementwise addition, 'cat' or 'concat': Concatenate along feature dimension.
        dim: Dimension of input tensors containing features.

    """
    if callable(arg):
        return arg

    if not isinstance(arg, str):
        raise TypeError("join_func() 'arg' must be str or callable")

    name = arg.lower()
    if name == "add":

        def add(args: Sequence[Tensor]) -> Tensor:
            assert args, "join_func('add') requires at least one input tensor"
            out = args[0]
            for i in range(1, len(args)):
                out = out + args[i]
            return out

        return add

    elif name in ("cat", "concat"):

        def cat(args: Sequence[Tensor]) -> Tensor:
            assert args, "join_func('cat') requires at least one input tensor"
            return torch.cat(args, dim=dim)

        return cat

    elif name == "mul":

        def mul(args: Sequence[Tensor]) -> Tensor:
            assert args, "join_func('mul') requires at least one input tensor"
            out = args[0]
            for i in range(1, len(args)):
                out = out * args[i]
            return out

        return mul

    raise ValueError("join_func() unknown merge function {name!r}")


class JoinLayer(LambdaLayer):
    r"""Merge network branches."""

    def __init__(self, arg: Union[str, JoinFunc], dim: int = 1) -> None:
        func = join_func(arg, dim=dim)
        super().__init__(func)

    def forward(self, xs: Sequence[Tensor]) -> Tensor:
        return self.func(xs)

    def extra_repr(self) -> str:
        return repr(self.func.__name__)



r"""Generic dataset classes and auxiliary functions."""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import copy as shallowcopy
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Iterable, Iterator
from typing import Mapping, Optional, Union, Sequence, TypeVar, overload

import pandas as pd

from torch.nn import Sequential
from torch.utils.data import Dataset as TorchDataset, Subset

from deepali.core.config import DataclassConfig
from deepali.core.typing import PathStr, Sample, is_namedtuple, is_path_str

from .transforms import Transform
from .transforms.image import ImageTransformConfig
from .transforms.image import image_transforms, prepend_read_image_transform


__all__ = (
    "Dataset",
    "MetaDataset",
    "ImageDataset",
    "ImageDatasetConfig",
    "GroupDataset",
    "JoinDataset",
    "read_table",
)


TDataset = TypeVar("TDataset", bound="Dataset")


class Dataset(TorchDataset, metaclass=ABCMeta):
    r"""Base class of datasets with optionally on-the-fly pre-processed samples.

    This map-style dataset base class is convenient for attaching data transformations to
    a given dataset. Otherwise, datasets may also derive directly from the respective
    ``torch.utils.data`` dataset classes or simply implement the expected interfaces.

    See also: https://pytorch.org/docs/stable/data.html

    """

    def __init__(self, transforms: Optional[Union[Transform, Sequence[Transform]]] = None):
        r"""Initialize dataset.

        If a dataset produces samples (i.e., a dictionary, named tuple, or custom dataclass)
        which contain fields with ``None`` values, ``collate_fn=collate_samples`` must be
        passed to ``torch.utils.data.DataLoader``. This custom collate function will ignore
        ``None`` values and pass these on to the respective batch entry. Auxiliary function
        ``prepare_batch()`` can be used to transfer the batch data retrieved by the data
        loader to the execution device.

        Args:
            transforms: Data preprocessing and augmentation transforms.
                If more than one transformation is given, these will be composed
                in the given order, where the first transformation in the sequence
                is applied first. When the data samples are passed directly to
                ``torch.utils.data.DataLoader``, the transformed sample data must
                be of type ``np.ndarray``, ``torch.Tensor``, or ``None``.

        """
        super().__init__()
        if transforms is None:
            transform = Sequential()
        elif isinstance(transforms, Sequential):
            transform = transforms
        else:
            if not isinstance(transforms, (list, tuple)):
                transforms = [transforms]
            transform = Sequential(*transforms)
        self._transform: Sequential = transform

    @abstractmethod
    def __len__(self) -> int:
        r"""Number of samples in dataset."""
        raise NotImplementedError

    def __getitem__(self, index: int) -> Sample:
        r"""Processed data of i-th dataset sample.

        Args:
            index: Index of dataset sample.

        Returns:
            Sample data.

        """
        sample = self.sample(index)
        sample = self._transform(sample)
        return sample

    @abstractmethod
    def sample(self, index: int) -> Sample:
        r"""Data of i-th dataset sample."""
        raise NotImplementedError

    def samples(self) -> Iterable[Sample]:
        r"""Get iterable over untransformed dataset samples."""

        class SampleIterator(object):
            def __init__(self, dataset: Dataset):
                self.dataset = dataset
                self.index = -1

            def __iter__(self) -> Iterator[Sample]:
                self.index = 0
                return self

            def __next__(self) -> Sample:
                if self.index >= len(self.dataset):
                    raise StopIteration
                sample = self.dataset.sample(self.index)
                self.index += 1
                return sample

        return SampleIterator(self)

    @overload
    def transform(self) -> Sequential:
        ...

    @overload
    def transform(
        self: TDataset,
        arg0: Union[Transform, Sequence[Transform], None],
        *args: Union[Transform, Sequence[Transform], None],
    ) -> TDataset:
        ...

    def transform(
        self: TDataset, *args: Union[Transform, Sequence[Transform], None]
    ) -> Union[Sequential, TDataset]:
        r"""Get composite data preprocessing and augmentation transform, or new dataset with specified transform."""
        if not args:
            return self._transform
        return shallowcopy(self).transform_(*args)

    def transform_(
        self: TDataset,
        arg0: Union[Transform, Sequence[Transform], None],
        *args: Union[Transform, Sequence[Transform], None],
    ) -> TDataset:
        r"""Set data preprocessing and augmentation transform of this dataset."""
        transforms = []
        for arg in [arg0, *args]:
            if arg is None:
                continue
            if isinstance(arg, (list, tuple)):
                transforms.extend(arg)
            else:
                transforms.append(arg)
        if not transforms:
            self._transform = None
        elif len(transforms) == 1 and isinstance(transforms[0], Sequential):
            self._transform = transforms[0]
        else:
            self._transform = Sequential(*transforms)
        return self

    @overload
    def transforms(self) -> List[Transform]:
        ...

    @overload
    def transforms(
        self: TDataset,
        arg0: Union[Transform, Sequence[Transform], None],
        *args: Union[Transform, Sequence[Transform], None],
    ) -> TDataset:
        ...

    def transforms(
        self: TDataset,
        *args: Union[Transform, Sequence[Transform], None],
    ) -> Union[List[Transform], TDataset]:
        r"""Get or set dataset transforms."""
        if not args:
            return [transform for transform in self._transform]
        return shallowcopy(self).transform_(*args)

    def transforms_(
        self,
        arg0: Union[Transform, Sequence[Transform], None],
        *args: Union[Transform, Sequence[Transform], None],
    ) -> Dataset:
        r"""Set data transforms of this dataset."""
        return self.transform_(arg0, *args)


class MetaDataset(Dataset):
    r"""Dataset of file path template strings and sample meta-data given by Pandas DataFrame.

    This dataset can be used in conjunction with data reader transforms to load the data from
    configured input file paths. For example, use the :class:`deepali.data.transforms.ReadImage`
    transform followed by image data preprocessing and augmentation functions for image data.
    The specified file path strings are Python format strings, where keywords are replaced by the
    respective column entries for the sample in the dataset index table (`pandas.DataFrame`).

    """

    def __init__(
        self,
        table: Union[Path, str, pd.DataFrame],
        paths: Optional[Mapping[str, Union[PathStr, Callable[..., PathStr]]]] = None,
        prefix: Optional[PathStr] = None,
        transforms: Optional[Union[Transform, Sequence[Transform]]] = None,
        **kwargs,
    ):
        r"""Initialize dataset.

        Args:
            table: Table with sample IDs, optionally sample specific input file path template
                strings (cf. ``paths``), and additional sample meta data.
            paths: File path template strings of input data files. The format string may contain keys ``prefix``,
                when a ``prefix`` path has been specified, and ``table`` column names. The dictionary keys of this
                argument are used as sample data dictionary keys for the respective file paths. When the path value
                is a string which matches exactly the name of a ``table`` column, the value of this column is used
                without configuring a file path template string. This is useful when the input ``table`` already
                specifies the file paths for each sample. Instead of a string, the dictionary value can be a
                callable function instead, which takes the ``table`` row values as keyword arguments, and must return
                the respectively formatted input file path string. When no ``paths`` are given, the dataset samples
                only contain the meta-data from the input ``table`` columns.
            prefix: Root directory of input file paths starting with ``"{prefix}/"``.
                If ``None`` and ``table`` is a file path, it is set to the directory containing the index table.
                Otherwise, template file path strings may not contain a ``{prefix}`` key if ``None``.
            transforms: Data preprocessing and augmentation transforms.
            kwargs: Additional format arguments used in addition to ``prefix`` and ``table`` column values.

        """
        if isinstance(table, (str, Path)):
            if prefix is None:
                path = Path(table).absolute()
                prefix = path.parent
            elif prefix:
                prefix = Path(prefix).absolute()
                path = prefix / Path(table)
            else:
                path = Path(table).absolute()
            table = read_table(path)
        if not isinstance(table, pd.DataFrame):
            raise TypeError(
                f"{type(self).__name__}() 'table' must be pandas.DataFrame or file path"
            )
        df: pd.DataFrame = table
        paths = {} if paths is None else dict(paths)
        if "meta" in df.columns:
            raise ValueError(
                f"{type(self).__name__} 'table' contains column with reserved name 'meta'"
            )
        if "meta" in paths:
            raise ValueError(f"{type(self).__name__} 'paths' contains reserved 'meta' key")
        prefix = Path(prefix).absolute() if prefix else None
        self.table = df
        self.paths = paths
        self.prefix = prefix
        self.kwargs = kwargs
        super().__init__(transforms=transforms)

    def __len__(self) -> int:
        r"""Number of samples in dataset."""
        return len(self.table)

    def row(self, index: int) -> Dict[str, Any]:
        r"""Get i-th table row values."""
        return self.table.iloc[index].to_dict()

    def sample(self, index: int) -> Dict[str, Any]:
        r"""Input file paths and/or meta-data of i-th sample in dataset."""
        meta = self.row(index)
        if not self.paths:
            return meta
        data = {}
        args = {"prefix": str(self.prefix)} if self.prefix else {}
        args["index"] = index
        args["index+1"] = index + 1
        args["index + 1"] = index + 1
        args.update(self.kwargs)
        args.update(meta)
        for name, path in self.paths.items():
            if callable(path):
                path = path(**args)
            elif path in meta:
                path = meta[path]
            else:
                path = path.format(**args)
            if not path:
                continue
            path = str(path)
            data[name] = path
            # Make path also available in meta-data dictionary such that even when data[name]
            # is replaced by the actual data stored at the given input file path (e.g., by a
            # ReadImage transform attached to the dataset), the file path remains available.
            meta[name] = path
        data["meta"] = meta
        return data

    def samples(self) -> Iterable[Dict[str, Any]]:
        r"""Get iterable over untransformed dataset samples."""

        class DatasetSampleIterator(object):
            def __init__(self, dataset: MetaDataset):
                self.dataset = dataset
                self.index = -1

            def __iter__(self) -> Iterator[Dict[str, Any]]:
                self.index = 0
                return self

            def __next__(self) -> Dict[str, Any]:
                if self.index >= len(self.dataset):
                    raise StopIteration
                sample = self.dataset.sample(self.index)
                self.index += 1
                return sample

        return DatasetSampleIterator(self)


@dataclass
class ImageDatasetConfig(DataclassConfig):
    r"""Configuration of image dataset."""

    table: PathStr
    prefix: Optional[PathStr] = None
    images: Mapping[str, PathStr] = field(default_factory=dict)
    transforms: Mapping[str, ImageTransformConfig] = field(default_factory=dict)

    @classmethod
    def _from_dict(
        cls, arg: Mapping[str, Any], parent: Optional[Path] = None
    ) -> ImageDatasetConfig:
        r"""Create configuration from dictionary.

        This function optionally re-organizes the dictionary entries to conform to the dataclass layout.
        It allows the image data transforms to be specified as separate "transforms" entry for each image.
        In this case, the image file path template string must given by the "path" dictionary entry.
        Additionally, a "read" image transform is added when a "dtype" or "device" is specified on which
        the image data is loaded and preprocessed can also be specified alongside the file "path".
        Any image "transforms" specified at the top-level are applied after any "transforms" specified
        underneath the "images" key.

        """
        arg = dict(arg)
        images = arg.pop("images", None) or {}
        transforms = arg.pop("transforms", None) or {}
        image_paths = {}
        for name, value in images.items():
            dtype = None
            device = None
            if isinstance(value, Mapping):
                if "path" not in value:
                    raise ValueError(
                        f"{cls.__name__}.from_dict() 'images' key '{name}' dict must contain 'path' entry"
                    )
                path = value["path"]
                dtype = value.get("dtype", dtype)
                device = value.get("device", device)
                image_transforms = value.get("transforms") or []
                if not isinstance(image_transforms, Sequence):
                    raise TypeError(
                        f"{cls.__name__}.from_dict() image 'transforms' value must be Sequence"
                    )
            elif is_path_str(value):
                path = Path(value).as_posix()
                image_transforms = []
            else:
                raise ValueError(
                    f"{cls.__name__}.from_dict() 'images' key '{name}' must be PathStr or dict with 'path' entry"
                )
            if name in transforms:
                item_transforms = transforms[name]
                if not isinstance(item_transforms, Sequence):
                    raise TypeError(
                        f"{cls.__name__}.from_dict() 'transforms' dict value must be Sequence"
                    )
                item_transforms = list(item_transforms)
            else:
                item_transforms = []
            image_transforms = image_transforms + item_transforms
            if dtype or device:
                image_transforms = prepend_read_image_transform(
                    image_transforms, dtype=dtype, device=device
                )
            transforms[name] = image_transforms
            image_paths[name] = path
        arg["images"] = {k: v for k, v in image_paths.items() if v}
        arg["transforms"] = {k: v for k, v in transforms.items() if v}
        return super()._from_dict(arg, parent)


class ImageDataset(MetaDataset):
    r"""Configurable image dataset."""

    @classmethod
    def from_config(cls, config: ImageDatasetConfig) -> ImageDataset:
        transforms = []
        for image_name in config.images:
            image_transforms_config = config.transforms.get(image_name, [])
            image_transforms_config = prepend_read_image_transform(image_transforms_config)
            item_transforms = image_transforms(image_transforms_config, key=image_name)
            transforms.extend(item_transforms)
        return cls(
            config.table,
            paths=config.images,
            prefix=config.prefix,
            transforms=transforms,
        )


class GroupDataset(TorchDataset):
    r"""Group samples in dataset."""

    def __init__(
        self,
        dataset: MetaDataset,
        groupby: Union[Sequence[str], str],
        sortby: Optional[Union[Sequence[str], str]] = None,
        ascending: bool = True,
    ) -> None:
        super().__init__()
        indices = []
        df = dataset.table
        if sortby:
            df = df.sort_values(sortby, ascending=ascending)
        groups = df.groupby(groupby)
        for _, group in groups:
            assert isinstance(group, pd.DataFrame)
            ilocs = [row[0] for row in group.itertuples(index=True)]
            indices.append(ilocs)
        self.dataset = dataset
        self.indices = indices

    def __len__(self) -> int:
        return len(self.indices)

    def __getitem__(self, index: int) -> Subset[Dict[str, Any]]:
        indices = self.indices[index]
        return Subset(self.dataset, indices)


class JoinDataset(Dataset):
    r"""Join dict entries from one or more datasets in a single dict."""

    def __init__(self, datasets: Iterable[Dataset]) -> None:
        super().__init__()
        datasets = list(datasets)
        if not all(len(dataset) == len(datasets[0]) for dataset in datasets):
            raise ValueError("JoinDataset() 'datasets' must have the same size")
        self.datasets = datasets

    def __len__(self) -> int:
        datasets = self.datasets
        return len(datasets[0]) if datasets else 0

    def sample(self, index: int) -> Sample:
        sample = {}
        for i, dataset in enumerate(self.datasets):
            data = dataset[index]
            if not isinstance(data, dict):
                if is_namedtuple(data):
                    data = data._asdict()
                else:
                    data = {str(i): data}
            for key, value in data.items():
                current = sample.get(key, None)
                if current is not None and current != value:
                    raise ValueError("JoinDataset() encountered ambiguous duplicate key '{key}'")
                sample[key] = value
        return sample


def read_table(path: PathStr) -> pd.DataFrame:
    r"""Read dataset index table."""
    path = Path(path).absolute()
    if path.suffix.lower() == ".h5":
        return pd.read_hdf(path)
    if path.suffix.lower() == ".tsv":
        return pd.read_csv(path, comment="#", skip_blank_lines=True, delimiter="\t")
    if path.suffix.lower() == ".csv":
        return pd.read_csv(path, comment="#", skip_blank_lines=True)
    raise NotImplementedError(f"read_table() does not support {path.suffix} file format")



r"""Regularization terms based on model parameters."""

from torch import Tensor

from .base import ParamsLoss


class L1Norm(ParamsLoss):
    r"""Regularization loss term based on L1-norm of model parameters."""

    def __init__(self, scale: float = 1e3) -> None:
        r"""Initialize loss term.

        Args:
            scale: Constant factor by which to scale loss value such that magnitude is in
                similar range to other registration loss terms, i.e., image matching terms.

        """
        super().__init__()
        self.scale = scale

    def forward(self, params: Tensor) -> Tensor:
        r"""Evaluate loss term for given model parameters."""
        return self.scale * params.abs().mean()


L1_Norm = L1Norm


class L2Norm(ParamsLoss):
    r"""Regularization loss term based on L2-norm of model parameters."""

    def __init__(self, scale: float = 1e6) -> None:
        r"""Initialize loss term.

        Args:
            scale: Constant factor by which to scale loss value such that magnitude is in
                similar range to other registration loss terms, i.e., image matching terms.

        """
        super().__init__()
        self.scale = scale

    def forward(self, params: Tensor) -> Tensor:
        r"""Evaluate loss term for given model parameters."""
        return self.scale * params.square().mean()


L2_Norm = L2Norm


class Sparsity(ParamsLoss):
    r"""Regularization loss term encouraging sparsity of non-zero model parameters."""

    def forward(self, params: Tensor) -> Tensor:
        r"""Evaluate loss term for given model parameters."""
        return params.abs().mean()



r"""Conversion functions between different representations of 3D rotations."""

# Copyright (C) 2017-2019, Arraiy, Inc., all rights reserved.
# Copyright (C) 2019-    , Kornia authors, all rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The conversion functions of this module have been copied and adapted from kornia
# https://github.com/kornia/kornia/blob/8b016621105fcbf1fb1ae13be8361471887bab16/kornia/geometry/conversions.py.
#
# Temporary support for QuaternionCoeffOrder.XYZW, which will be deprecated in kornia >0.6,
# has been removed from the functions here (cf. https://github.com/kornia/kornia/issues/903).


import torch
import torch.nn.functional as F


__all__ = (
    "angle_axis_to_rotation_matrix",
    "angle_axis_to_quaternion",
    "normalize_quaternion",
    "rotation_matrix_to_angle_axis",
    "rotation_matrix_to_quaternion",
    "quaternion_to_angle_axis",
    "quaternion_to_rotation_matrix",
    "quaternion_log_to_exp",
    "quaternion_exp_to_log",
)


def angle_axis_to_rotation_matrix(angle_axis: torch.Tensor) -> torch.Tensor:
    r"""Convert 3d vector of axis-angle rotation to 3x3 rotation matrix

    Args:
        angle_axis (torch.Tensor): tensor of 3d vector of axis-angle rotations.

    Returns:
        torch.Tensor: tensor of 3x3 rotation matrices.

    Shape:
        - Input: :math:`(N, 3)`
        - Output: :math:`(N, 3, 3)`

    Example:
        >>> input = torch.rand(1, 3)  # Nx3
        >>> output = angle_axis_to_rotation_matrix(input)  # Nx3x3
    """
    if not isinstance(angle_axis, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(type(angle_axis)))

    if not angle_axis.shape[-1] == 3:
        raise ValueError("Input size must be a (*, 3) tensor. Got {}".format(angle_axis.shape))

    def _compute_rotation_matrix(angle_axis, theta2, eps=1e-6):
        # We want to be careful to only evaluate the square root if the
        # norm of the angle_axis vector is greater than zero. Otherwise
        # we get a division by zero.
        k_one = 1.0
        theta = torch.sqrt(theta2)
        wxyz = angle_axis / (theta + eps)
        wx, wy, wz = torch.chunk(wxyz, 3, dim=1)
        cos_theta = torch.cos(theta)
        sin_theta = torch.sin(theta)

        r00 = cos_theta + wx * wx * (k_one - cos_theta)
        r10 = wz * sin_theta + wx * wy * (k_one - cos_theta)
        r20 = -wy * sin_theta + wx * wz * (k_one - cos_theta)
        r01 = wx * wy * (k_one - cos_theta) - wz * sin_theta
        r11 = cos_theta + wy * wy * (k_one - cos_theta)
        r21 = wx * sin_theta + wy * wz * (k_one - cos_theta)
        r02 = wy * sin_theta + wx * wz * (k_one - cos_theta)
        r12 = -wx * sin_theta + wy * wz * (k_one - cos_theta)
        r22 = cos_theta + wz * wz * (k_one - cos_theta)
        rotation_matrix = torch.cat([r00, r01, r02, r10, r11, r12, r20, r21, r22], dim=1)
        return rotation_matrix.view(-1, 3, 3)

    def _compute_rotation_matrix_taylor(angle_axis):
        rx, ry, rz = torch.chunk(angle_axis, 3, dim=1)
        k_one = torch.ones_like(rx)
        rotation_matrix = torch.cat([k_one, -rz, ry, rz, k_one, -rx, -ry, rx, k_one], dim=1)
        return rotation_matrix.view(-1, 3, 3)

    # stolen from ceres/rotation.h

    _angle_axis = torch.unsqueeze(angle_axis, dim=1)
    theta2 = torch.matmul(_angle_axis, _angle_axis.transpose(1, 2))
    theta2 = torch.squeeze(theta2, dim=1)

    # compute rotation matrices
    rotation_matrix_normal = _compute_rotation_matrix(angle_axis, theta2)
    rotation_matrix_taylor = _compute_rotation_matrix_taylor(angle_axis)

    # create mask to handle both cases
    eps = 1e-6
    mask = (theta2 > eps).view(-1, 1, 1).to(theta2.device)
    mask_pos = (mask).type_as(theta2)
    mask_neg = (mask == False).type_as(theta2)  # noqa

    # create output pose matrix
    batch_size = angle_axis.shape[0]
    rotation_matrix = torch.eye(3).to(angle_axis.device).type_as(angle_axis)
    rotation_matrix = rotation_matrix.view(1, 3, 3).repeat(batch_size, 1, 1)
    # fill output matrix with masked values
    rotation_matrix[..., :3, :3] = (
        mask_pos * rotation_matrix_normal + mask_neg * rotation_matrix_taylor
    )
    return rotation_matrix  # Nx3x3


def rotation_matrix_to_angle_axis(rotation_matrix: torch.Tensor) -> torch.Tensor:
    r"""Convert 3x3 rotation matrix to Rodrigues vector.

    Args:
        rotation_matrix (torch.Tensor): rotation matrix.

    Returns:
        torch.Tensor: Rodrigues vector transformation.

    Shape:
        - Input: :math:`(N, 3, 3)`
        - Output: :math:`(N, 3)`

    Example:
        >>> input = torch.rand(2, 3, 3)  # Nx3x3
        >>> output = rotation_matrix_to_angle_axis(input)  # Nx3
    """
    if not isinstance(rotation_matrix, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(rotation_matrix)}")

    if not rotation_matrix.shape[-2:] == (3, 3):
        raise ValueError(f"Input size must be a (*, 3, 3) tensor. Got {rotation_matrix.shape}")
    quaternion: torch.Tensor = rotation_matrix_to_quaternion(rotation_matrix)
    return quaternion_to_angle_axis(quaternion)


def rotation_matrix_to_quaternion(
    rotation_matrix: torch.Tensor, eps: float = 1.0e-8
) -> torch.Tensor:
    r"""Convert 3x3 rotation matrix to 4d quaternion vector.

    The quaternion vector has components in (w, x, y, z) or (x, y, z, w) format.

    .. note::
        The (x, y, z, w) order is going to be deprecated in favor of efficiency.

    Args:
        rotation_matrix (torch.Tensor): the rotation matrix to convert.
        eps (float): small value to avoid zero division. Default: 1e-8.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: the rotation in quaternion.

    Shape:
        - Input: :math:`(*, 3, 3)`
        - Output: :math:`(*, 4)`

    Example:
        >>> input = torch.rand(4, 3, 3)  # Nx3x3
        >>> output = rotation_matrix_to_quaternion(input, eps=torch.finfo(input.dtype).eps,
        ...                                        order=QuaternionCoeffOrder.WXYZ)  # Nx4
    """
    if not isinstance(rotation_matrix, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(rotation_matrix)}")

    if not rotation_matrix.shape[-2:] == (3, 3):
        raise ValueError(f"Input size must be a (*, 3, 3) tensor. Got {rotation_matrix.shape}")

    def safe_zero_division(numerator: torch.Tensor, denominator: torch.Tensor) -> torch.Tensor:
        eps: float = torch.finfo(numerator.dtype).tiny  # type: ignore
        return numerator / torch.clamp(denominator, min=eps)

    rotation_matrix_vec: torch.Tensor = rotation_matrix.view(*rotation_matrix.shape[:-2], 9)

    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.chunk(rotation_matrix_vec, chunks=9, dim=-1)

    trace: torch.Tensor = m00 + m11 + m22

    def trace_positive_cond():
        sq = torch.sqrt(trace + 1.0) * 2.0  # sq = 4 * qw.
        qw = 0.25 * sq
        qx = safe_zero_division(m21 - m12, sq)
        qy = safe_zero_division(m02 - m20, sq)
        qz = safe_zero_division(m10 - m01, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_1():
        sq = torch.sqrt(1.0 + m00 - m11 - m22 + eps) * 2.0  # sq = 4 * qx.
        qw = safe_zero_division(m21 - m12, sq)
        qx = 0.25 * sq
        qy = safe_zero_division(m01 + m10, sq)
        qz = safe_zero_division(m02 + m20, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_2():
        sq = torch.sqrt(1.0 + m11 - m00 - m22 + eps) * 2.0  # sq = 4 * qy.
        qw = safe_zero_division(m02 - m20, sq)
        qx = safe_zero_division(m01 + m10, sq)
        qy = 0.25 * sq
        qz = safe_zero_division(m12 + m21, sq)
        return torch.cat((qw, qx, qy, qz), dim=-1)

    def cond_3():
        sq = torch.sqrt(1.0 + m22 - m00 - m11 + eps) * 2.0  # sq = 4 * qz.
        qw = safe_zero_division(m10 - m01, sq)
        qx = safe_zero_division(m02 + m20, sq)
        qy = safe_zero_division(m12 + m21, sq)
        qz = 0.25 * sq
        return torch.cat((qw, qx, qy, qz), dim=-1)

    where_2 = torch.where(m11 > m22, cond_2(), cond_3())
    where_1 = torch.where((m00 > m11) & (m00 > m22), cond_1(), where_2)

    quaternion: torch.Tensor = torch.where(trace > 0.0, trace_positive_cond(), where_1)
    return quaternion


def normalize_quaternion(quaternion: torch.Tensor, eps: float = 1.0e-12) -> torch.Tensor:
    r"""Normalizes a quaternion.

    The quaternion should be in (x, y, z, w) format.

    Args:
        quaternion (torch.Tensor): a tensor containing a quaternion to be
          normalized. The tensor can be of shape :math:`(*, 4)`.
        eps (Optional[bool]): small value to avoid division by zero.
          Default: 1e-12.

    Return:
        torch.Tensor: the normalized quaternion of shape :math:`(*, 4)`.

    Example:
        >>> quaternion = torch.tensor((1., 0., 1., 0.))
        >>> normalize_quaternion(quaternion)
        tensor([0.7071, 0.0000, 0.7071, 0.0000])
    """
    if not isinstance(quaternion, torch.Tensor):
        raise TypeError("Input type is not a torch.Tensor. Got {}".format(type(quaternion)))

    if not quaternion.shape[-1] == 4:
        raise ValueError("Input must be a tensor of shape (*, 4). Got {}".format(quaternion.shape))
    return F.normalize(quaternion, p=2.0, dim=-1, eps=eps)


# based on:
# https://github.com/matthew-brett/transforms3d/blob/8965c48401d9e8e66b6a8c37c65f2fc200a076fa/transforms3d/quaternions.py#L101
# https://github.com/tensorflow/graphics/blob/master/tensorflow_graphics/geometry/transformation/rotation_matrix_3d.py#L247


def quaternion_to_rotation_matrix(quaternion: torch.Tensor) -> torch.Tensor:
    r"""Converts a quaternion to a rotation matrix.

    The quaternion should be in (x, y, z, w) or (w, x, y, z) format.

    Args:
        quaternion (torch.Tensor): a tensor containing a quaternion to be
          converted. The tensor can be of shape :math:`(*, 4)`.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: the rotation matrix of shape :math:`(*, 3, 3)`.

    Example:
        >>> quaternion = torch.tensor((0., 0., 0., 1.))
        >>> quaternion_to_rotation_matrix(quaternion, order=QuaternionCoeffOrder.WXYZ)
        tensor([[-1.,  0.,  0.],
                [ 0., -1.,  0.],
                [ 0.,  0.,  1.]])
    """
    if not isinstance(quaternion, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(quaternion)}")

    if not quaternion.shape[-1] == 4:
        raise ValueError(f"Input must be a tensor of shape (*, 4). Got {quaternion.shape}")

    # normalize the input quaternion
    quaternion_norm: torch.Tensor = normalize_quaternion(quaternion)

    # unpack the normalized quaternion components
    w, x, y, z = torch.chunk(quaternion_norm, chunks=4, dim=-1)

    # compute the actual conversion
    tx: torch.Tensor = 2.0 * x
    ty: torch.Tensor = 2.0 * y
    tz: torch.Tensor = 2.0 * z
    twx: torch.Tensor = tx * w
    twy: torch.Tensor = ty * w
    twz: torch.Tensor = tz * w
    txx: torch.Tensor = tx * x
    txy: torch.Tensor = ty * x
    txz: torch.Tensor = tz * x
    tyy: torch.Tensor = ty * y
    tyz: torch.Tensor = tz * y
    tzz: torch.Tensor = tz * z
    one: torch.Tensor = torch.tensor(1.0)

    matrix: torch.Tensor = torch.stack(
        (
            one - (tyy + tzz),
            txy - twz,
            txz + twy,
            txy + twz,
            one - (txx + tzz),
            tyz - twx,
            txz - twy,
            tyz + twx,
            one - (txx + tyy),
        ),
        dim=-1,
    ).view(-1, 3, 3)

    if len(quaternion.shape) == 1:
        matrix = torch.squeeze(matrix, dim=0)
    return matrix


def quaternion_to_angle_axis(quaternion: torch.Tensor) -> torch.Tensor:
    """Convert quaternion vector to angle axis of rotation.

    The quaternion should be in (x, y, z, w) or (w, x, y, z) format.

    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h

    Args:
        quaternion (torch.Tensor): tensor with quaternions.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: tensor with angle axis of rotation.

    Shape:
        - Input: :math:`(*, 4)` where `*` means, any number of dimensions
        - Output: :math:`(*, 3)`

    Example:
        >>> quaternion = torch.rand(2, 4)  # Nx4
        >>> angle_axis = quaternion_to_angle_axis(quaternion)  # Nx3
    """
    if not torch.is_tensor(quaternion):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(quaternion)}")

    if not quaternion.shape[-1] == 4:
        raise ValueError(f"Input must be a tensor of shape Nx4 or 4. Got {quaternion.shape}")

    # unpack input and compute conversion
    q1: torch.Tensor = torch.tensor([])
    q2: torch.Tensor = torch.tensor([])
    q3: torch.Tensor = torch.tensor([])
    cos_theta: torch.Tensor = torch.tensor([])

    cos_theta = quaternion[..., 0]
    q1 = quaternion[..., 1]
    q2 = quaternion[..., 2]
    q3 = quaternion[..., 3]

    sin_squared_theta: torch.Tensor = q1 * q1 + q2 * q2 + q3 * q3

    sin_theta: torch.Tensor = torch.sqrt(sin_squared_theta)
    two_theta: torch.Tensor = 2.0 * torch.where(
        cos_theta < 0.0,
        torch.atan2(-sin_theta, -cos_theta),
        torch.atan2(sin_theta, cos_theta),
    )

    k_pos: torch.Tensor = two_theta / sin_theta
    k_neg: torch.Tensor = 2.0 * torch.ones_like(sin_theta)
    k: torch.Tensor = torch.where(sin_squared_theta > 0.0, k_pos, k_neg)

    angle_axis: torch.Tensor = torch.zeros_like(quaternion)[..., :3]
    angle_axis[..., 0] += q1 * k
    angle_axis[..., 1] += q2 * k
    angle_axis[..., 2] += q3 * k
    return angle_axis


def quaternion_log_to_exp(quaternion: torch.Tensor, eps: float = 1.0e-8) -> torch.Tensor:
    r"""Applies exponential map to log quaternion.

    The quaternion should be in (x, y, z, w) or (w, x, y, z) format.

    Args:
        quaternion (torch.Tensor): a tensor containing a quaternion to be
          converted. The tensor can be of shape :math:`(*, 3)`.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: the quaternion exponential map of shape :math:`(*, 4)`.

    Example:
        >>> quaternion = torch.tensor((0., 0., 0.))
        >>> quaternion_log_to_exp(quaternion, eps=torch.finfo(quaternion.dtype).eps,
        ...                       order=QuaternionCoeffOrder.WXYZ)
        tensor([1., 0., 0., 0.])
    """
    if not isinstance(quaternion, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(quaternion)}")

    if not quaternion.shape[-1] == 3:
        raise ValueError(f"Input must be a tensor of shape (*, 3). Got {quaternion.shape}")

    # compute quaternion norm
    norm_q: torch.Tensor = torch.norm(quaternion, p=2, dim=-1, keepdim=True).clamp(min=eps)

    # compute scalar and vector
    quaternion_vector: torch.Tensor = quaternion * torch.sin(norm_q) / norm_q
    quaternion_scalar: torch.Tensor = torch.cos(norm_q)

    # compose quaternion and return
    quaternion_exp: torch.Tensor = torch.tensor([])
    quaternion_exp = torch.cat((quaternion_scalar, quaternion_vector), dim=-1)

    return quaternion_exp


def quaternion_exp_to_log(quaternion: torch.Tensor, eps: float = 1.0e-8) -> torch.Tensor:
    r"""Applies the log map to a quaternion.

    The quaternion should be in (x, y, z, w) format.

    Args:
        quaternion (torch.Tensor): a tensor containing a quaternion to be
          converted. The tensor can be of shape :math:`(*, 4)`.
        eps (float): A small number for clamping.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: the quaternion log map of shape :math:`(*, 3)`.

    Example:
        >>> quaternion = torch.tensor((1., 0., 0., 0.))
        >>> quaternion_exp_to_log(quaternion, eps=torch.finfo(quaternion.dtype).eps,
        ...                       order=QuaternionCoeffOrder.WXYZ)
        tensor([0., 0., 0.])
    """
    if not isinstance(quaternion, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(quaternion)}")

    if not quaternion.shape[-1] == 4:
        raise ValueError(f"Input must be a tensor of shape (*, 4). Got {quaternion.shape}")

    # unpack quaternion vector and scalar
    quaternion_vector: torch.Tensor = torch.tensor([])
    quaternion_scalar: torch.Tensor = torch.tensor([])

    quaternion_scalar = quaternion[..., 0:1]
    quaternion_vector = quaternion[..., 1:4]

    # compute quaternion norm
    norm_q: torch.Tensor = torch.norm(quaternion_vector, p=2, dim=-1, keepdim=True).clamp(min=eps)

    # apply log map
    quaternion_log: torch.Tensor = (
        quaternion_vector * torch.acos(torch.clamp(quaternion_scalar, min=-1.0, max=1.0)) / norm_q
    )

    return quaternion_log


# based on:
# https://github.com/facebookresearch/QuaterNet/blob/master/common/quaternion.py#L138


def angle_axis_to_quaternion(angle_axis: torch.Tensor) -> torch.Tensor:
    r"""Convert an angle axis to a quaternion.

    The quaternion vector has components in (x, y, z, w) or (w, x, y, z) format.

    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h

    Args:
        angle_axis (torch.Tensor): tensor with angle axis.
        order (QuaternionCoeffOrder): quaternion coefficient order. Default: 'xyzw'.
          Note: 'xyzw' will be deprecated in favor of 'wxyz'.

    Return:
        torch.Tensor: tensor with quaternion.

    Shape:
        - Input: :math:`(*, 3)` where `*` means, any number of dimensions
        - Output: :math:`(*, 4)`

    Example:
        >>> angle_axis = torch.rand(2, 3)  # Nx3
        >>> quaternion = angle_axis_to_quaternion(angle_axis, order=QuaternionCoeffOrder.WXYZ)  # Nx4
    """
    if not torch.is_tensor(angle_axis):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(angle_axis)}")

    if not angle_axis.shape[-1] == 3:
        raise ValueError(f"Input must be a tensor of shape Nx3 or 3. Got {angle_axis.shape}")

    # unpack input and compute conversion
    a0: torch.Tensor = angle_axis[..., 0:1]
    a1: torch.Tensor = angle_axis[..., 1:2]
    a2: torch.Tensor = angle_axis[..., 2:3]
    theta_squared: torch.Tensor = a0 * a0 + a1 * a1 + a2 * a2

    theta: torch.Tensor = torch.sqrt(theta_squared)
    half_theta: torch.Tensor = theta * 0.5

    mask: torch.Tensor = theta_squared > 0.0
    ones: torch.Tensor = torch.ones_like(half_theta)

    k_neg: torch.Tensor = 0.5 * ones
    k_pos: torch.Tensor = torch.sin(half_theta) / theta
    k: torch.Tensor = torch.where(mask, k_pos, k_neg)
    w: torch.Tensor = torch.where(mask, torch.cos(half_theta), ones)

    quaternion: torch.Tensor = torch.zeros(
        size=(*angle_axis.shape[:-1], 4),
        dtype=angle_axis.dtype,
        device=angle_axis.device,
    )
    quaternion[..., 1:2] = a0 * k
    quaternion[..., 2:3] = a1 * k
    quaternion[..., 3:4] = a2 * k
    quaternion[..., 0:1] = w
    return quaternion



import torch

from deepali.core import Grid
import deepali.core.functional as U
import deepali.losses.functional as L


def test_losses_flow_bending() -> None:
    data = torch.randn((1, 2, 16, 24))
    a = L.bending_energy(data, mode="bspline", stride=1, reduction="none")
    b = L.bspline_bending_energy(data, stride=1, reduction="none")
    assert torch.allclose(a, b)


def test_losses_flow_curvature() -> None:
    grid = Grid(size=(16, 14))
    offset = U.translation([0.1, 0.2]).unsqueeze_(0)
    flow = U.affine_flow(offset, grid)
    flow.requires_grad = True
    loss = L.curvature_loss(flow)
    assert loss.requires_grad
    loss = loss.detach()
    assert loss.abs().max().lt(1e-5)



r"""Utility functions and classes for building applications.

Currently, these utilities are primarily intended for use in example applications and tutorials
of this library. It is not recommended to build external applications on these while it is in a
non-stable release version (<1.0).

.. note::
    These auxiliary modules should only depend on the :mod:`.core` library and third-party packages.

"""
